# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import deejayd.interfaces
from deejayd.interfaces import DeejaydError,\
                               DeejaydAnswer,\
                               DeejaydKeyValue, DeejaydList,\
                               DeejaydFileList,\
                               DeejaydMediaList, DeejaydDvdInfo
from deejayd.ui.config import DeejaydConfig
from deejayd import mediafilters, player, sources, mediadb, database, plugins

# Exception imports
import deejayd.sources
import deejayd.mediadb.library


def require_mode(mode_name):
    def require_mode_instance(func):

        def require_mode_func(self, *__args, **__kw):
            if self.sources.is_available(mode_name):
                return func(self, *__args, **__kw)
            else:
                raise DeejaydError(_("mode %s is not activated.") % mode_name)

        require_mode_func.__name__ = func.__name__
        return require_mode_func

    return require_mode_instance

# This decorator is used for the core interface to provide :
# - A simple way of converting python types to deejayd answers as required by
#   the defined network interface. This is to ensure that the core API stays
#   the same as the client library API. This way, code written as a deejayd
#   client can become a full media player for free by swaping
#   deejayd.net.DeejaydClient and deejayd.core.DeejaydCore .
# - A simple way to make this optionnal using an optionnal objanswer argument.
#   Please don't use objanswer=False if you think that what you write may one
#   day be used as a deejayd network client. This was originally provided for
#   use in the deejayd.net.commandsXML module.
def returns_deejaydanswer(answer_class):
    def returns_deejaydanswer_instance(func):

        def interface_clean_func(*__args, **__kw):
            if __kw.has_key('objanswer'):
                objanswer = __kw['objanswer']
                del __kw['objanswer']
            else:
                objanswer = True
            if objanswer:
                ans = answer_class()
                try:
                    res = func(*__args, **__kw)
                except DeejaydError, txt:
                    ans.set_error(txt)
                else:
                    if res == None:
                        ans.contents = True
                    elif answer_class == DeejaydMediaList:
                        if isinstance(res, tuple):
                            ans.set_filter(res[1])
                            ans.set_sort(res[2])
                            res = res[0]
                        ans.set_medias(res)
                    elif answer_class == DeejaydFileList:
                        root_dir, dirs, files = res
                        if root_dir != None:
                            ans.set_rootdir(root_dir)
                        ans.set_files(files)
                        ans.set_directories(dirs)
                    elif answer_class == DeejaydDvdInfo:
                        ans.set_dvd_content(res)
                    else:
                        ans.contents = res
                return ans
            else:
                return func(*__args, **__kw)

        interface_clean_func.__name__ = func.__name__
        return interface_clean_func

    return returns_deejaydanswer_instance


class DeejaydStaticPlaylist(deejayd.interfaces.DeejaydStaticPlaylist):

    def __init__(self, deejaydcore, pl_id, name):
        self.deejaydcore = deejaydcore
        self.db, self.library = deejaydcore.db, deejaydcore.audio_library
        self.name = name
        self.pl_id = pl_id

    @returns_deejaydanswer(DeejaydMediaList)
    def get(self, first=0, length=-1):
        songs = self.db.get_static_medialist(self.pl_id,\
            infos = self.library.media_attr)
        last = length == -1 and len(songs) or int(first) + int(length)
        return songs[int(first):last]

    @returns_deejaydanswer(DeejaydAnswer)
    def add_paths(self, paths):
        ids = []
        for path in paths:
            try: medias = self.library.get_all_files(path)
            except deejayd.mediadb.library.NotFoundException:
                try: medias = self.library.get_file(path)
                except deejayd.mediadb.library.NotFoundException:
                    raise DeejaydError(_('Path %s not found in library') % path)
            for m in medias: ids.append(m["media_id"])
        self.add_songs(ids)

    @returns_deejaydanswer(DeejaydAnswer)
    def add_songs(self, song_ids):
        self.db.add_to_static_medialist(self.pl_id, song_ids)
        self.db.connection.commit()
        self.deejaydcore._dispatch_signame('playlist.update',\
                {"pl_id": self.pl_id})


class DeejaydMagicPlaylist(deejayd.interfaces.DeejaydMagicPlaylist):
    """ Magic playlist object """

    def __init__(self, deejaydcore, pl_id, name):
        self.deejaydcore = deejaydcore
        self.db, self.library = deejaydcore.db, deejaydcore.audio_library
        self.name = name
        self.pl_id = pl_id

    @returns_deejaydanswer(DeejaydMediaList)
    def get(self, first=0, length=-1):
        properties = dict(self.db.get_magic_medialist_properties(self.pl_id))
        if properties["use-or-filter"] == "1":
            filter = mediafilters.Or()
        else:
            filter = mediafilters.And()
        if properties["use-limit"] == "1":
            sort = [(properties["limit-sort-value"],\
                     properties["limit-sort-direction"])]
            limit = int(properties["limit-value"])
        else:
            sort, limit = [], None
        filter.filterlist = self.db.get_magic_medialist_filters(self.pl_id)
        songs = self.library.search(filter, sort, limit)
        last = length == -1 and len(songs) or int(first) + int(length)
        return (songs[int(first):last], filter, None)

    @returns_deejaydanswer(DeejaydAnswer)
    def add_filter(self, filter):
        if filter.type != "basic":
            raise DeejaydError(\
                    _("Only basic filters are allowed for magic playlist"))
        self.db.add_magic_medialist_filters(self.pl_id, [filter])
        self.deejaydcore._dispatch_signame('playlist.update',\
                {"pl_id": self.pl_id})

    @returns_deejaydanswer(DeejaydAnswer)
    def remove_filter(self, filter):
        record_filters = self.db.get_magic_medialist_filters(self.pl_id)
        new_filters = []
        for record_filter in record_filters:
            if not filter.equals(record_filter):
                new_filters.append(record_filter)
        self.db.set_magic_medialist_filters(self.name, new_filters)
        self.deejaydcore._dispatch_signame('playlist.update',\
                {"pl_id": self.pl_id})

    @returns_deejaydanswer(DeejaydAnswer)
    def clear_filters(self):
        self.db.set_magic_medialist_filters(self.name, [])
        self.deejaydcore._dispatch_signame('playlist.update',\
                {"pl_id": self.pl_id})

    @returns_deejaydanswer(DeejaydKeyValue)
    def get_properties(self):
        return dict(self.db.get_magic_medialist_properties(self.pl_id))

    @returns_deejaydanswer(DeejaydAnswer)
    def set_property(self, key, value):
        self.db.set_magic_medialist_property(self.pl_id, key, value)
        self.deejaydcore._dispatch_signame('playlist.update',\
                {"pl_id": self.pl_id})


class DeejaydWebradioList(deejayd.interfaces.DeejaydWebradioList):

    def __init__(self, deejaydcore):
        self.deejaydcore = deejaydcore
        self.source = self.deejaydcore.sources.get_source('webradio')

    @returns_deejaydanswer(DeejaydMediaList)
    def get(self, first = 0, length = -1):
        wrs = self.source.get_content()
        last = length == -1 and len(wrs) or int(first) + int(length)
        return wrs[int(first):last]

    @returns_deejaydanswer(DeejaydKeyValue)
    def get_available_sources(self):
        return dict(self.source.get_available_sources())

    @returns_deejaydanswer(DeejaydList)
    def get_source_categories(self, source_name):
        return self.source.get_source_categories(source_name)

    @returns_deejaydanswer(DeejaydAnswer)
    def set_source(self, source_name):
        self.source.set_source(source_name)

    @returns_deejaydanswer(DeejaydAnswer)
    def set_source_categorie(self, categorie):
        self.source.set_source_categorie(categorie)

    @returns_deejaydanswer(DeejaydAnswer)
    def add_webradio(self, name, urls):
        self.source.add(urls, name)

    @returns_deejaydanswer(DeejaydAnswer)
    def delete_webradios(self, wr_ids):
        ids = map(int, wr_ids)
        self.source.delete(ids)

    @returns_deejaydanswer(DeejaydAnswer)
    def clear(self):
        self.source.clear()


class DeejaydQueue(deejayd.interfaces.DeejaydQueue):

    def __init__(self, deejaydcore):
        self.deejaydcore = deejaydcore
        self.source = self.deejaydcore.sources.get_source('queue')

    @returns_deejaydanswer(DeejaydMediaList)
    def get(self, first = 0, length = -1):
        songs = self.source.get_content()
        last = length == -1 and len(songs) or int(first) + int(length)
        return songs[int(first):last]

    @returns_deejaydanswer(DeejaydAnswer)
    def add_songs(self, song_ids, pos = None):
        p = pos and int(pos) or None
        try: self.source.add_song(song_ids, pos = p)
        except deejayd.sources._base.SourceError, ex:
            raise DeejaydError(str(ex))

    @returns_deejaydanswer(DeejaydAnswer)
    def add_paths(self, paths, pos = None):
        p = pos and int(pos) or None
        try: self.source.add_path(paths, p)
        except deejayd.sources._base.SourceError, ex:
            raise DeejaydError(str(ex))

    @returns_deejaydanswer(DeejaydAnswer)
    def load_playlists(self, pl_ids, pos=None):
        pos = pos and int(pos) or None
        try: self.source.load_playlist(pl_ids, pos)
        except deejayd.sources._base.SourceError, ex:
            raise DeejaydError(str(ex))

    @returns_deejaydanswer(DeejaydAnswer)
    def move(self, ids, new_pos):
        ids = [int(id) for id in ids]
        try: self.source.move(ids, new_pos)
        except deejayd.sources._base.SourceError, ex:
            raise DeejaydError(str(ex))

    @returns_deejaydanswer(DeejaydAnswer)
    def clear(self):
        self.source.clear()

    @returns_deejaydanswer(DeejaydAnswer)
    def del_songs(self, ids):
        ids = [int(id) for id in ids]
        try: self.source.delete(ids)
        except deejayd.sources._base.SourceError, ex:
            raise DeejaydError(str(ex))


class DeejaydPlaylistMode(deejayd.interfaces.DeejaydPlaylistMode):
    """Audio playlist mode."""

    def __init__(self, deejaydcore):
        self.deejaydcore = deejaydcore
        self.source = self.deejaydcore.sources.get_source("playlist")

    @returns_deejaydanswer(DeejaydMediaList)
    def get(self, first=0, length=-1):
        songs = self.source.get_content()
        last = length == -1 and len(songs) or int(first) + int(length)
        return songs[int(first):last]

    @returns_deejaydanswer(DeejaydKeyValue)
    def save(self, name):
        if name == "":
            raise DeejaydError(_("Set a playlist name"))
        return self.source.save(name)

    @returns_deejaydanswer(DeejaydAnswer)
    def add_paths(self, paths, pos=None):
        p = pos and int(pos) or None
        try: self.source.add_path(paths, pos = p)
        except deejayd.sources._base.SourceError, ex:
            raise DeejaydError(str(ex))

    @returns_deejaydanswer(DeejaydAnswer)
    def add_songs(self, song_ids, pos=None):
        p = pos and int(pos) or None
        try: self.source.add_song(song_ids, pos = p)
        except deejayd.sources._base.SourceError, ex:
            raise DeejaydError(str(ex))

    @returns_deejaydanswer(DeejaydAnswer)
    def loads(self, pl_ids, pos=None):
        pos = pos and int(pos) or None
        try: self.source.load_playlist(pl_ids, pos)
        except deejayd.sources._base.SourceError, ex:
            raise DeejaydError(str(ex))

    @returns_deejaydanswer(DeejaydAnswer)
    def move(self, ids, new_pos):
        ids = [int(id) for id in ids]
        try: self.source.move(ids, int(new_pos))
        except deejayd.sources._base.SourceError, ex:
            raise DeejaydError(str(ex))

    @returns_deejaydanswer(DeejaydAnswer)
    def shuffle(self):
        try: self.source.shuffle()
        except deejayd.sources._base.SourceError, ex:
            raise DeejaydError(str(ex))

    @returns_deejaydanswer(DeejaydAnswer)
    def clear(self):
        try: self.source.clear()
        except deejayd.sources._base.SourceError, ex:
            raise DeejaydError(str(ex))

    @returns_deejaydanswer(DeejaydAnswer)
    def del_songs(self, ids):
        ids = [int(id) for id in ids]
        try: self.source.delete(ids)
        except deejayd.sources._base.SourceError, ex:
            raise DeejaydError(str(ex))


class DeejaydPanel(deejayd.interfaces.DeejaydPanel):

    def __init__(self, deejaydcore):
        self.deejaydcore = deejaydcore
        self.source = self.deejaydcore.sources.get_source("panel")

    @returns_deejaydanswer(DeejaydMediaList)
    def get(self, first=0, length=-1):
        songs, filters, sort = self.source.get_content()
        last = length == -1 and len(songs) or int(first) + int(length)
        return (songs[int(first):last], filters, sort)

    @returns_deejaydanswer(DeejaydList)
    def get_panel_tags(self):
        return self.source.get_panel_tags()

    @returns_deejaydanswer(DeejaydKeyValue)
    def get_active_list(self):
        return self.source.get_active_list()

    @returns_deejaydanswer(DeejaydAnswer)
    def set_active_list(self, type, pl_id=""):
        try: self.source.set_active_list(type, pl_id)
        except TypeError:
            raise DeejaydError(_("Not supported type"))
        except deejayd.sources._base.SourceError, ex:
            raise DeejaydError(str(ex))

    @returns_deejaydanswer(DeejaydAnswer)
    def set_panel_filters(self, tag, values):
        try: self.source.set_panel_filters(tag, values)
        except deejayd.sources._base.SourceError, ex:
            raise DeejaydError(str(ex))

    @returns_deejaydanswer(DeejaydAnswer)
    def remove_panel_filters(self, tag):
        try: self.source.remove_panel_filters(tag)
        except deejayd.sources._base.SourceError, ex:
            raise DeejaydError(str(ex))

    @returns_deejaydanswer(DeejaydAnswer)
    def clear_panel_filters(self):
        self.source.clear_panel_filters()

    @returns_deejaydanswer(DeejaydAnswer)
    def set_search_filter(self, tag, value):
        try: self.source.set_search_filter(tag, value)
        except deejayd.sources._base.SourceError, ex:
            raise DeejaydError(str(ex))

    @returns_deejaydanswer(DeejaydAnswer)
    def clear_search_filter(self):
        self.source.clear_search_filter()

    @returns_deejaydanswer(DeejaydAnswer)
    def set_sorts(self, sorts):
        try: self.source.set_sorts(sorts)
        except deejayd.sources._base.SourceError, ex:
            raise DeejaydError(str(ex))

class DeejaydVideo(deejayd.interfaces.DeejaydVideo):
    """Video mode."""

    def __init__(self, deejaydcore):
        self.deejaydcore = deejaydcore
        self.source = self.deejaydcore.sources.get_source('video')

    @returns_deejaydanswer(DeejaydMediaList)
    def get(self, first = 0, length = -1):
        videos, filters, sort = self.source.get_content()
        last = length == -1 and len(videos) or int(first) + int(length)
        return (videos[int(first):last], filters, sort)

    @returns_deejaydanswer(DeejaydAnswer)
    def set(self, value, type = "directory"):
        try: self.source.set(type, value)
        except deejayd.sources._base.SourceError, ex:
            raise DeejaydError(str(ex))

    @returns_deejaydanswer(DeejaydAnswer)
    def set_sorts(self, sorts):
        try: self.source.set_sorts(sorts)
        except deejayd.sources._base.SourceError, ex:
            raise DeejaydError(str(ex))

class DeejayDaemonCore(deejayd.interfaces.DeejaydCore):

    def __init__(self, config=None):
        deejayd.interfaces.DeejaydCore.__init__(self)

        if not config:
            config = DeejaydConfig()

        self.db = database.init(config)
        self.plugin_manager = plugins.PluginManager(config)

        self.player = player.init(self.db, self.plugin_manager, config)
        self.player.register_dispatcher(self)

        self.audio_library,self.video_library, self.watcher = \
            mediadb.init(self.db, self.player,config)
        self.audio_library.register_dispatcher(self)
        if self.video_library:
            self.video_library.register_dispatcher(self)

        self.sources = sources.init(self.player, self.db, self.audio_library,
                                    self.video_library, self.plugin_manager,
                                    config)
        self.sources.register_dispatcher(self)
        for source in self.sources.sources_obj.values():
            source.register_dispatcher(self)

        if not self.db.structure_created:
            self.update_audio_library(objanswer=False)
            if self.video_library:
                self.update_video_library(objanswer=False)

        # start inotify thread when we are sure that all init stuff are ok
        if self.watcher:
            self.watcher.start()

    def close(self):
        for obj in (self.watcher,self.player,self.sources,self.audio_library,\
                    self.video_library,self.db):
            if obj != None: obj.close()

    @returns_deejaydanswer(DeejaydAnswer)
    def play_toggle(self):
        if self.player.get_state() == player._base.PLAYER_PLAY:
            current_media = self.player.get_playing()
            if current_media['type'] == 'webradio':
                # There is no point in pausing radio streams.
                try: self.player.stop()
                except player.PlayerError, err: raise DeejaydError(err)
            else:
                self.player.pause()
        else:
            try: self.player.play()
            except player.PlayerError, err:
                raise DeejaydError(err)

    @returns_deejaydanswer(DeejaydAnswer)
    def stop(self):
        try: self.player.stop()
        except player.PlayerError, err:
            raise DeejaydError(err)

    @returns_deejaydanswer(DeejaydAnswer)
    def previous(self):
        try: self.player.previous()
        except player.PlayerError, err:
            raise DeejaydError(err)

    @returns_deejaydanswer(DeejaydAnswer)
    def next(self):
        try: self.player.next()
        except player.PlayerError, err:
            raise DeejaydError(err)

    @returns_deejaydanswer(DeejaydAnswer)
    def seek(self, pos, relative = False):
        self.player.set_position(int(pos), relative)

    @returns_deejaydanswer(DeejaydMediaList)
    def get_current(self):
        medias = []
        current = self.player.get_playing()
        if current != None:
            medias.append(current)
        return medias

    @require_mode("playlist")
    def get_playlist(self):
        return DeejaydPlaylistMode(self)

    @require_mode("panel")
    def get_panel(self):
        return DeejaydPanel(self)

    @require_mode("webradio")
    def get_webradios(self):
        return DeejaydWebradioList(self)

    @require_mode("video")
    def get_video(self):
        return DeejaydVideo(self)

    def get_queue(self):
        return DeejaydQueue(self)

    @returns_deejaydanswer(DeejaydAnswer)
    def go_to(self, id, id_type = "id", source = None):
        if id_type not in ("dvd_id","track","chapter","id","pos"):
            raise DeejaydError(_("Bad value for id_type parm"))
        if id_type != "dvd_id":
            try: id = int(id)
            except ValueError:
                raise DeejaydError(_("Bad value for id parm"))

        try: self.player.go_to(id, id_type, source)
        except player.PlayerError, err:
            raise DeejaydError(err)

    @returns_deejaydanswer(DeejaydAnswer)
    def set_volume(self, volume_value):
        self.player.set_volume(int(volume_value))

    @returns_deejaydanswer(DeejaydAnswer)
    def set_option(self, source, option_name, option_value):
        try: self.sources.set_option(source, option_name, option_value)
        except sources.UnknownSourceException:
            raise DeejaydError(_('Mode %s not supported') % source)
        except sources._base.SourceError, ex:
            raise DeejaydError(str(ex))

    @returns_deejaydanswer(DeejaydAnswer)
    def set_mode(self, mode_name):
        try: self.sources.set_source(mode_name)
        except sources.UnknownSourceException:
            raise DeejaydError(_('Mode %s not supported') % mode_name)

    @returns_deejaydanswer(DeejaydKeyValue)
    def get_mode(self):
        av_sources = self.sources.get_available_sources()
        modes = {}
        for s in self.sources.get_all_sources():
            modes[s] = s in av_sources
        return modes

    @returns_deejaydanswer(DeejaydAnswer)
    def set_player_option(self, name, value):
        if name != "aspect_ratio":
            try: value = int(value)
            except (ValueError,TypeError):
                raise DeejaydError(_("Param value is not an int"))

        try: self.player.set_option(name, value)
        except KeyError:
            raise DeejaydError(_("Option %s does not exist") % name)
        except NotImplementedError:
            raise DeejaydError(_("Option %s is not supported for this backend")\
                                % name)
        except player.PlayerError, err:
            raise DeejaydError(err)

    @returns_deejaydanswer(DeejaydKeyValue)
    def get_status(self):
        status = self.player.get_status()
        status.extend(self.sources.get_status())
        status.extend(self.audio_library.get_status())
        if self.video_library:
            status.extend(self.video_library.get_status())
        return dict(status)

    @returns_deejaydanswer(DeejaydKeyValue)
    def get_stats(self):
        ans = self.db.get_stats()
        return dict(ans)

    @returns_deejaydanswer(DeejaydKeyValue)
    def update_audio_library(self, force = False, sync = False):
        return {'audio_updating_db': self.audio_library.update(force, sync)}

    @require_mode("video")
    @returns_deejaydanswer(DeejaydKeyValue)
    def update_video_library(self, force = False, sync = False):
        if not self.video_library:
            raise DeejaydError(_("Video mode disabled"))
        return {'video_updating_db': self.video_library.update(force, sync)}

    @returns_deejaydanswer(DeejaydKeyValue)
    def create_recorded_playlist(self, name, type):
        if name == "":
            raise DeejaydError(_("Set a playlist name"))
        # first search if this pls already exist
        try: self.db.get_medialist_id(name, type)
        except ValueError: pass
        else: # pls already exists
            raise DeejaydError(_("This playlist already exists"))

        if type == "static":
            pl_id = self.db.set_static_medialist(name, [])
        elif type == "magic":
            pl_id = self.db.set_magic_medialist_filters(name, [])
            pl = DeejaydMagicPlaylist(self, pl_id, name)
            # set default properties for this playlist
            default = {
                    "use-or-filter": "0",
                    "use-limit": "0",
                    "limit-value": "50",
                    "limit-sort-value": "title",
                    "limit-sort-direction": "ascending"
                    }
            for (k, v) in default.items():
                pl.set_property(k, v)
        self._dispatch_signame('playlist.listupdate')
        return {"pl_id": pl_id, "name": name, "type": type}

    def get_recorded_playlist(self, id, name = "", type = "static"):
        try: pl_id, name, type = self.db.is_medialist_exists(id)
        except TypeError:
            raise DeejaydError(_("Playlist with id %s not found.") % str(id))
        if type == "static":
            return DeejaydStaticPlaylist(self, pl_id, name)
        elif type == "magic":
            return DeejaydMagicPlaylist(self, pl_id, name)

    @returns_deejaydanswer(DeejaydAnswer)
    def erase_playlist(self, ids):
        for id in ids:
            self.db.delete_medialist(id)
            self._dispatch_signame('playlist.listupdate')

    @returns_deejaydanswer(DeejaydMediaList)
    def get_playlist_list(self):
        return [{"name": pl, "id":id, "type":type}\
            for (id, pl, type) in self.db.get_medialist_list() if not \
            pl.startswith("__") or not pl.endswith("__")]

    @returns_deejaydanswer(DeejaydAnswer)
    def set_media_rating(self, media_ids, rating, type = "audio"):
        if int(rating) not in range(0, 5):
            raise DeejaydError(_("Bad rating value"))

        try: library = getattr(self, type+"_library")
        except AttributeError:
            raise DeejaydError(_('Type %s is not supported') % (type,))
        for id in media_ids:
            try: library.set_file_info(int(id), "rating", rating)
            except TypeError:
                raise DeejaydError(_("%s library not activated") % type)
            except deejayd.mediadb.library.NotFoundException:
                raise DeejaydError(_("File with id %s not found") % str(id))

    @returns_deejaydanswer(DeejaydFileList)
    def get_audio_dir(self,dir = None):
        if dir == None: dir = ""
        try: contents = self.audio_library.get_dir_content(dir)
        except deejayd.mediadb.library.NotFoundException:
            raise DeejaydError(_('Directory %s not found in database') % dir)

        return dir, contents['dirs'], contents['files']

    @returns_deejaydanswer(DeejaydKeyValue)
    def get_audio_cover(self,media_id):
        try: cover = self.audio_library.get_cover(media_id)
        except deejayd.mediadb.library.NotFoundException:
            raise DeejaydError(_('Cover not found'))
        return cover

    @returns_deejaydanswer(DeejaydMediaList)
    def audio_search(self, pattern, type = 'all'):
        if type not in ('all','title','genre','filename','artist','album'):
            raise DeejaydError(_('Type %s is not supported') % (type,))
        if type == "all":
            filter = mediafilters.Or()
            for tag in ('title','genre','artist','album'):
                filter.combine(mediafilters.Contains(tag, pattern))
        else:
            filter = mediafilters.Contains(type, pattern)
        songs = self.audio_library.search(filter,\
                mediafilters.DEFAULT_AUDIO_SORT)

        return songs

    @require_mode("video")
    @returns_deejaydanswer(DeejaydFileList)
    def get_video_dir(self,dir = None):
        if not self.video_library:
            raise DeejaydError(_("Video mode disabled"))

        if dir == None: dir = ""
        try: contents = self.video_library.get_dir_content(dir)
        except deejayd.mediadb.library.NotFoundException:
            raise DeejaydError(_('Directory %s not found in database') % dir)

        return dir, contents['dirs'], contents['files']

    @require_mode("dvd")
    @returns_deejaydanswer(DeejaydAnswer)
    def dvd_reload(self):
        try: self.sources.get_source("dvd").load()
        except sources.dvd.DvdError, msg:
            raise DeejaydError('%s' % msg)

    @require_mode("dvd")
    @returns_deejaydanswer(DeejaydDvdInfo)
    def get_dvd_content(self):
        return self.sources.get_source("dvd").get_content()

    @returns_deejaydanswer(DeejaydList)
    def mediadb_list(self, tag, filter):
        return [x[0] for x in self.db.list_tags(tag, filter)]


# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


import signal, sys, locale

from twisted.python import log
from deejayd.ui.config import DeejaydConfig

ERROR = 0
INFO = 1
DEBUG = 2

level = DeejaydConfig().get("general","log")
log_level = {"error": ERROR, "info": INFO, \
             "debug": DEBUG}[level]


class LogFile:

    def __init__(self, path, autosignal=True):
        self.path = path
        self.open()
        if autosignal:
            self.set_reopen_signal()

    def open(self):
        try:
            self.fd = open(self.path, 'a')
        except IOError:
            sys.exit("Unable to open the log file %s" % self.path)

    def reopen(self):
        self.fd.close()
        self.open()

    def __reopen_cb(self, signal, frame):
        self.reopen()

    def set_reopen_signal(self, sig=signal.SIGHUP, callback=None):
        if not callback:
            callback = self.__reopen_cb
        signal.signal(sig, callback)


class SignaledFileLogObserver(log.FileLogObserver):

    def __init__(self, path):
        self.log_file = LogFile(path, False)
        self.log_file.open()
        self.__observe_log_file()
        self.log_file.set_reopen_signal(callback=self.__reopen_cb)

    def __observe_log_file(self):
        log.FileLogObserver.__init__(self, self.log_file.fd)

    def __reopen_cb(self, signal, frame):
        self.stop()
        self.log_file.reopen()
        self.__observe_log_file()
        self.start()


def __log(log_msg):
    try: log.msg(log_msg.encode(locale.getpreferredencoding()))
    except UnicodeError:
        # perharps prefered encoding not correctly set, force to UTF-8
        log.msg(log_msg.encode('utf-8'))
    
def err(err, fatal = False):
    msg = _("ERROR - %s") % err
    __log(msg)
    if fatal: sys.exit(err)

def msg(msg):
    __log(msg)

def info(msg):
    if log_level >= INFO:
        msg = _("INFO - %s") % msg
        __log(msg)

def debug(msg):
    if log_level >= DEBUG:
        msg = _("DEBUG - %s") % msg
        __log(msg)

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from deejayd.database import schema

def upgrade(cursor, backend, config):
    # get recorded webradio
    cursor.execute("SELECT * FROM webradio")
    webradios = cursor.fetchall()
    # drop old webradio table and create new one
    cursor.execute("DROP TABLE webradio")
    for table in schema.db_schema:
        if table.name in ("webradio", "webradio_entries"):
            for stmt in backend.to_sql(table):
                cursor.execute(stmt)
    # record webradio in new table
    for wid, name, url in webradios:
        cursor.execute("SELECT id from webradio WHERE name=%s", (name[:-2],))
        try:
            (id,) = cursor.fetchone()
        except (TypeError, ValueError):
            cursor.execute("INSERT INTO webradio (name)VALUES(%s)",\
                           (name[:-2],))
            id = cursor.lastrowid
        cursor.execute("INSERT INTO webradio_entries \
            (webradio_id,url)VALUES(%s,%s)", (id, url))

    # remove useless state entries
    query = "DELETE FROM variables WHERE name = %s"
    cursor.executemany(query, [
                ("playlist-playorder",),
                ("video-playorder",),
                ("panel-playorder",),
                ("queue-playorder",),
                ("playlist-repeat",),
                ("video-repeat",),
                ("panel-repeat",),
                ("queueid",),
                ("playlistid",),
                ("panelid",),
                ("webradioid",),
                ("videoid",),
                ("dvdid",),
                ("panel-type",),
                ("panel-value",),
            ])

    # update db version
    sql = [
        "UPDATE variables SET value = '13' WHERE name = 'database_version';",
        ]
    for s in sql:
        cursor.execute(s)

# vim ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


import sys, ctypes


try:
    _xinelib = ctypes.cdll.LoadLibrary('libxine.so.1')
except (ImportError, OSError), e:
    raise ImportError, e


# void xine_get_version (int *major, int *minor, int *sub)
_xinelib.xine_get_version.argstype = (ctypes.POINTER(ctypes.c_int),
                                      ctypes.POINTER(ctypes.c_int),
                                      ctypes.POINTER(ctypes.c_int),)

# int  xine_check_version (int major, int minor, int sub)
_xinelib.xine_check_version.argtypes = (ctypes.c_int, ctypes.c_int,
                                        ctypes.c_int)
_xinelib.xine_check_version.restype = ctypes.c_int

# char *xine_get_file_extensions (xine_t *self)
_xinelib.xine_get_file_extensions.argtypes = (ctypes.c_void_p, )
_xinelib.xine_get_file_extensions.restype = ctypes.c_char_p

# char *const *xine_list_input_plugins(xine_t *self)
_xinelib.xine_list_input_plugins.argtypes = (ctypes.c_void_p, )
_xinelib.xine_list_input_plugins.restype = ctypes.POINTER(ctypes.c_char_p)

# xine_t *xine_new (void)
_xinelib.xine_new.restype = ctypes.c_void_p

# void xine_config_load  (xine_t *self, const char *cfg_filename)
_xinelib.xine_config_load.argstype = (ctypes.c_void_p, ctypes.c_char_p)

# const char *xine_get_homedir(void)
_xinelib.xine_get_homedir.restype = ctypes.c_char_p

# void xine_init (xine_t *self)
_xinelib.xine_init.argstype = (ctypes.c_void_p, )

# void xine_exit (xine_t *self)
_xinelib.xine_exit.argstype = (ctypes.c_void_p, )

# void xine_engine_set_param(xine_t *self, int param, int value)
_xinelib.xine_engine_set_param.argstype = (ctypes.c_void_p,
                                           ctypes.c_int, ctypes.c_int, )

# xine_audio_port_t *xine_open_audio_driver (xine_t *self, const char *id,
#                       void *data)
_xinelib.xine_open_audio_driver.argstype = (ctypes.c_void_p,
                                            ctypes.c_char_p, ctypes.c_void_p, )
_xinelib.xine_open_audio_driver.restype = ctypes.c_void_p

# xine_video_port_t *xine_open_video_driver (xine_t *self, const char *id,
#                       int visual, void *data)
_xinelib.xine_open_video_driver.argstype = (ctypes.c_void_p,
                                            ctypes.c_char_p,
                                            ctypes.c_int, ctypes.c_void_p, )
_xinelib.xine_open_video_driver.restype = ctypes.c_void_p

# void xine_close_audio_driver (xine_t *self, xine_audio_port_t  *driver)
_xinelib.xine_close_audio_driver.argstype = (ctypes.c_void_p, ctypes.c_void_p, )

# void xine_close_video_driver (xine_t *self, xine_video_port_t  *driver)
_xinelib.xine_close_video_driver.argstype = (ctypes.c_void_p, ctypes.c_void_p, )

# int    xine_port_send_gui_data (xine_video_port_t *vo,
#                    int type, void *data)
_xinelib.xine_port_send_gui_data.argstype = (ctypes.c_void_p,
                                             ctypes.c_int, ctypes.c_void_p, )
_xinelib.xine_port_send_gui_data.restype = ctypes.c_int

# xine_stream_t *xine_stream_new (xine_t *self,
#                xine_audio_port_t *ao, xine_video_port_t *vo)
_xinelib.xine_stream_new.argstype = (ctypes.c_void_p,
                                     ctypes.c_void_p, ctypes.c_void_p,)
_xinelib.xine_stream_new.restype = ctypes.c_void_p

# int xine_open (xine_stream_t *stream, const char *mrl)
_xinelib.xine_open.argstype = (ctypes.c_void_p, ctypes.c_char_p, )
_xinelib.xine_open.restype = ctypes.c_int

# int  xine_play (xine_stream_t *stream, int start_pos, int start_time)
_xinelib.xine_play.argstype = (ctypes.c_void_p, ctypes.c_int, ctypes.c_int, )
_xinelib.xine_play.restype = ctypes.c_int

# void xine_stop (xine_stream_t *stream)
_xinelib.xine_stop.argstype = (ctypes.c_void_p, )

_xinelib.xine_usec_sleep.argtypes = (ctypes.c_int, )

# void xine_close (xine_stream_t *stream)
_xinelib.xine_close.argstype = (ctypes.c_void_p, )

# void xine_dispose (xine_stream_t *stream)
_xinelib.xine_dispose.argstype = (ctypes.c_void_p, )

class x11_visual_t(ctypes.Structure):
    _fields_ = (
        ('display', ctypes.c_void_p),
        ('screen', ctypes.c_int),
        ('d', ctypes.c_ulong), # Drawable
        ('user_data', ctypes.c_void_p),
        ('dest_size_cb', ctypes.c_void_p),
        ('frame_output_cb', ctypes.c_void_p),
        ('lock_display', ctypes.c_void_p),
        ('unlock_display', ctypes.c_void_p),
    )

# dest size callback
xine_dest_size_cb = ctypes.CFUNCTYPE(ctypes.c_void_p, ctypes.c_void_p,
                     ctypes.c_int, ctypes.c_int, ctypes.c_double,
                     ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int),
                     ctypes.POINTER(ctypes.c_double))

# frame output callback
xine_frame_output_cb = ctypes.CFUNCTYPE(ctypes.c_void_p, ctypes.c_void_p,
                  ctypes.c_int, ctypes.c_int, ctypes.c_double,
                  ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int),
                  ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int),
                  ctypes.POINTER(ctypes.c_double),
                  ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int))

class xine_event_t(ctypes.Structure):
    _fields_ = (
        ('type', ctypes.c_int),
        ('stream', ctypes.c_void_p),
        ('data', ctypes.c_void_p),
        ('data_length', ctypes.c_int),
    )

class xine_ui_message_data_t(ctypes.Structure):
    _fields_ = (
        ('compatibility_num_buttons', ctypes.c_int),
        ('compatibility_str_len', ctypes.c_int),
        ('compatibility_str', 256 * ctypes.c_char),
        ('type', ctypes.c_int),
        ('explanation', ctypes.c_int),
        ('num_parameters', ctypes.c_int),
        ('parameters', ctypes.c_int),
        ('messages', ctypes.c_char),
    )

# event listener callback type
xine_event_listener_cb_t = ctypes.CFUNCTYPE(ctypes.c_void_p, ctypes.c_void_p,
                                            ctypes.POINTER(xine_event_t))

# void xine_event_create_listener_thread(xine_event_queue_t *queue,
#    xine_event_listener_cb_t callback,
#    void *user_data)
_xinelib.xine_event_create_listener_thread.argtypes = (ctypes.c_void_p,
                                                       ctypes.c_void_p,
                                                       ctypes.c_void_p)

# xine_event_queue_t *xine_event_new_queue(xine_stream_t *stream)
_xinelib.xine_event_new_queue.argtypes = (ctypes.c_void_p, )
_xinelib.xine_event_new_queue.restype = ctypes.c_void_p

# void xine_event_dispose_queue(xine_event_queue_t *queue)
_xinelib.xine_event_dispose_queue.argtypes = (ctypes.c_void_p, )

# void xine_set_param (xine_stream_t *stream, int param, int value)
_xinelib.xine_set_param.argtypes = (ctypes.c_void_p, ctypes.c_int, ctypes.c_int)

# int xine_get_param (xine_stream_t *stream, int param)
_xinelib.xine_get_param.argtypes = (ctypes.c_void_p, ctypes.c_int)
_xinelib.xine_get_param.restype = ctypes.c_int

# char *xine_get_meta_info(xine_stream_t *stream, int info)
_xinelib.xine_get_meta_info.argtypes = (ctypes.c_void_p, ctypes.c_int)
_xinelib.xine_get_meta_info.restype = ctypes.c_char_p

# int xine_get_stream_info(xine_stream_t *stream, int info)
_xinelib.xine_get_stream_info.argtypes = (ctypes.c_void_p, ctypes.c_int)
_xinelib.xine_get_stream_info.restype = ctypes.c_int

# int xine_get_status (xine_stream_t *stream)
_xinelib.xine_get_status.argtypes = (ctypes.c_void_p, )
_xinelib.xine_get_status.restype = ctypes.c_int

# int xine_get_pos_length (xine_stream_t *stream, int *pos_stream,
#                          int *pos_time, int *length_time)
_xinelib.xine_get_pos_length.argtypes = (ctypes.c_void_p,
                                         ctypes.POINTER(ctypes.c_int),
                                         ctypes.POINTER(ctypes.c_int),
                                         ctypes.POINTER(ctypes.c_int))

# int xine_get_status (xine_stream_t *stream)
_xinelib.xine_get_status.argtypes = (ctypes.c_void_p, )
_xinelib.xine_get_status.restype = ctypes.c_int

# int xine_get_audio_lang(xine_stream_t *stream, int channel, char *lang)
_xinelib.xine_get_audio_lang.restype = ctypes.c_int
_xinelib.xine_get_audio_lang.argtypes = (ctypes.c_void_p, ctypes.c_int,
                                         ctypes.c_char_p)

# int xine_get_spu_lang(xine_stream_t *stream, int channel, char *lang)
_xinelib.xine_get_spu_lang.restype = ctypes.c_int
_xinelib.xine_get_spu_lang.argtypes = (ctypes.c_void_p, ctypes.c_int,
                                       ctypes.c_char_p)

# xine_osd_t *xine_osd_new(xine_stream_t *self, int x, int y,
#                          int width, int height)
_xinelib.xine_osd_new.argtypes = (ctypes.c_void_p, ctypes.c_int, ctypes.c_int,
                                  ctypes.c_int, ctypes.c_int)
_xinelib.xine_osd_new.restype = ctypes.c_void_p

# void xine_osd_free(xine_osd_t *self)
_xinelib.xine_osd_free.argtypes = (ctypes.c_void_p, )

# uint32_t xine_osd_get_capabilities(xine_osd_t *self)
_xinelib.xine_osd_get_capabilities.restype = ctypes.c_int
_xinelib.xine_osd_get_capabilities.argtypes = (ctypes.c_void_p, )

# void xine_osd_set_text_palette(xine_osd_t *self,int palette_number,
#                                int color_base )
_xinelib.xine_osd_set_text_palette.argtypes = (ctypes.c_void_p, ctypes.c_int,
                                               ctypes.c_int)

# int xine_osd_set_font(xine_osd_t *self, const char *fontname, int size)
_xinelib.xine_osd_set_font.restype = ctypes.c_int
_xinelib.xine_osd_set_font.argtypes = (ctypes.c_void_p, ctypes.c_char_p,
                                       ctypes.c_int)

# void xine_osd_set_position(xine_osd_t *self, int x, int y)
_xinelib.xine_osd_set_position.argtypes = (ctypes.c_void_p, ctypes.c_int,
                                        ctypes.c_int)

# void xine_osd_draw_text(xine_osd_t *self, int x1, int y1, char *text,
#                         int color_base)
_xinelib.xine_osd_draw_text.argtypes = (ctypes.c_void_p, ctypes.c_int,
                                        ctypes.c_int, ctypes.c_char_p,
                                        ctypes.c_int)

# void xine_osd_show(xine_osd_t *self, int64_t vpts)
_xinelib.xine_osd_show.argtypes = (ctypes.c_void_p, ctypes.c_int)

# void  xine_osd_show_unscaled (xine_osd_t *self, int64_t vpts)
_xinelib.xine_osd_show_unscaled.argtypes = (ctypes.c_void_p, ctypes.c_int)

# void  xine_osd_hide(xine_osd_t *self, int64_t vpts)
_xinelib.xine_osd_hide.argtypes = (ctypes.c_void_p, ctypes.c_int)

# void  xine_osd_clear(xine_osd_t *self)
_xinelib.xine_osd_clear.argtypes = (ctypes.c_void_p, )


# copy functions from the library
module = sys.modules[__name__]
for name in dir(_xinelib):
    if name.startswith('xine_'):
        setattr(module, name, getattr(_xinelib, name))


# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


import os, unittest

from testdeejayd import TestCaseWithData

from deejayd.mediafilters import *

import deejayd.database
import deejayd.ui.config
from deejayd.database.dbobjects import SQLizer


class TestDatabase(TestCaseWithData):

    def setUp(self):
        super(TestDatabase, self).setUp()

        config = deejayd.ui.config.DeejaydConfig()
        config.set('database', 'db_type', 'sqlite')
        dbfilename = self.testdata.getRandomString() + '.db'
        self.dbfilename = os.path.join('/tmp', dbfilename)
        config.set('database', 'db_name', self.dbfilename)
        self.db = deejayd.database.init(config)

        self.sqlizer = SQLizer()

    def tearDown(self):
        self.db.close()
        os.unlink(self.dbfilename)

    def test_basicfilter_save_retrieve(self):
        """Test the saving and retrieval of a basic filter"""
        f_class = self.testdata.getRandomElement(BASIC_FILTERS)
        f = f_class(self.testdata.getRandomString(),
                    self.testdata.getRandomString())
        sqlf = self.sqlizer.translate(f)

        fid = sqlf.save(self.db)
        self.db.connection.commit()

        cursor = self.db.connection.cursor()
        retrieved_filter = self.db.get_filter(cursor, fid)

        self.failUnless(f.equals(retrieved_filter))

        sqlf.tag = self.testdata.getRandomString()
        sqlf.save(self.db)
        self.db.connection.commit()

        retrieved_filter = self.db.get_filter(cursor, fid)
        self.failUnless(sqlf.equals(retrieved_filter))

        cursor.close()

    def test_complex_filter_save_retrieve(self):
        """Test the saving and retrieval of a complex filter"""
        f = self.testdata.get_sample_filter()
        sqlf = self.sqlizer.translate(f)

        fid = sqlf.save(self.db)
        self.db.connection.commit()

        cursor = self.db.connection.cursor()
        retrieved_filter = self.db.get_filter(cursor, fid)

        self.assert_filter_matches_sample(retrieved_filter)
        cursor.close()

    def test_magic_medialist_save_retrieve(self):
        """Test the saving and retrieval of a magic medialist"""
        f = self.testdata.get_sample_filter()

        pl_name = self.testdata.getRandomString()
        pl_id = self.db.set_magic_medialist_filters(pl_name, [f])

        # retrieve magic playlist
        retrieved_filters = self.db.get_magic_medialist_filters(pl_id)
        self.assertEqual(len(retrieved_filters), 1)
        self.failUnless(f.equals(retrieved_filters[0]))

        # resave the same playlist
        self.db.set_magic_medialist_filters(pl_name, retrieved_filters)

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


__all__ = (
            'BASIC_FILTERS', 'NAME2BASIC',
            'Equals', 'NotEquals', 'Contains', 'NotContains', 'Regexi',
            'Higher', 'Lower',
            'COMPLEX_FILTERS', 'NAME2COMPLEX',
            'And', 'Or',
            "DEFAULT_AUDIO_SORT", "DEFAULT_VIDEO_SORT"
          )


class MediaFilter(object):

    def get_identifier(self):
        return self.__class__.__name__.lower()

    def get_name(self):
        return self.__class__.__name__.lower()

    def __str__(self):
        return NotImplementedError


class BasicFilter(MediaFilter):
    type = 'basic'

    def __init__(self, tag, pattern):
        super(BasicFilter, self).__init__()
        self.tag = tag
        self.pattern = pattern

    def equals(self, filter):
        if filter.type == 'basic' and filter.get_name() == self.get_name():
            return filter.tag == self.tag and filter.pattern == self.pattern
        return False

    def __str__(self):
        return self.repr_str % (self.tag, self.pattern)


class Equals(BasicFilter):    repr_str = "(%s == '%s')"
class NotEquals(BasicFilter): repr_str = "(%s != '%s')"
class Contains(BasicFilter):  repr_str = "(%s == '%%%s%%')"
class NotContains(BasicFilter):  repr_str = "(%s != '%%%s%%')"
class Regexi(BasicFilter):    repr_str = "(%s ~ /%s/)"
class Higher(BasicFilter):    repr_str = "(%s >= %s)"
class Lower(BasicFilter):    repr_str = "(%s <= %s)"


BASIC_FILTERS = (
                  Equals,
                  NotEquals,
                  Contains,
                  NotContains,
                  Regexi,
                  Higher,
                  Lower,
                )

NAME2BASIC = dict([(x(None, None).get_identifier(), x) for x in BASIC_FILTERS])


class ComplexFilter(MediaFilter):
    type = 'complex'

    def __init__(*__args, **__kwargs):
        self = __args[0]
        super(ComplexFilter, self).__init__()

        self.filterlist = []
        for filter in __args[1:]:
            self.combine(filter)

    def combine(self, filter):
        self.filterlist.append(filter)

    def equals(self, filter):
        if filter.type == 'complex' and len(filter) == len(self.filterlist):
            for ft in self.filterlist:
                if not filter.find_filter(ft): return False
            return True
        return False

    def find_filter_by_tag(self, tag):
        return [ft for ft in self.filterlist\
            if ft.type == 'basic' and ft.tag == tag]

    def find_filter_by_name(self, name):
        return [ft for ft in self.filterlist if ft.get_name() == name]

    def find_filter(self, filter):
        return [ft for ft in self.filterlist if ft.equals(filter)]

    def remove_filter(self, filter):
        filters = self.find_filter(filter)
        if not filters:
            raise ValueError
        for ft in filters: self.filterlist.remove(ft)

    def __iter__(self):
        return iter(self.filterlist)

    def __getitem__(self, idx):
        return self.filterlist[idx]

    def __len__(self):
        return len(self.filterlist)

    def __str__(self):
        filters_str = map(str, self.filterlist)
        return "(%s)" % self.repr_joiner.join(filters_str)


class And(ComplexFilter): repr_joiner = " && "
class Or(ComplexFilter):  repr_joiner = " || "


COMPLEX_FILTERS = (
                    And,
                    Or,
                  )

NAME2COMPLEX = dict([(x().get_identifier(), x) for x in COMPLEX_FILTERS])

DEFAULT_AUDIO_SORT = [("album", "ascending"), ("discnumber", "ascending"),\
        ("tracknumber", "ascending")]
DEFAULT_VIDEO_SORT = [("title", "ascending")]

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""Deejayd Client library testing"""
import threading

from testdeejayd import TestCaseWithServer

from testdeejayd.server import TestServer
from testdeejayd.coreinterface import InterfaceTests, InterfaceSubscribeTests
from deejayd.net.client import DeejayDaemonSync, DeejayDaemonAsync, \
                               DeejayDaemonHTTP, DeejaydError, \
                               DeejaydWebradioList


class TestHTTPClient(TestCaseWithServer):
    """Test the http client library"""

    def setUp(self):
        TestCaseWithServer.setUp(self)

        # Instanciate the server object of the client library
        self.deejayd = DeejayDaemonHTTP('localhost', self.webServerPort)

    def testPing(self):
        """Ping server with a http connection"""
        self.failUnless(self.deejayd.ping().get_contents())


class TestSyncClient(TestCaseWithServer, InterfaceTests):
    """Test the DeejaydClient library in synchroneous mode."""

    def setUp(self):
        TestCaseWithServer.setUp(self)
        # Instanciate the server object of the client library
        self.deejayd = DeejayDaemonSync()
        self.deejayd.connect('localhost', self.serverPort)

    def tearDown(self):
        self.deejayd.disconnect()
        TestCaseWithServer.tearDown(self)

    def testPing(self):
        """Ping server"""
        self.failUnless(self.deejayd.ping().get_contents())


class TestAsyncClient(TestCaseWithServer, InterfaceSubscribeTests):
    """Test the DeejaydClient library in asynchroenous mode."""

    def setUp(self):
        TestCaseWithServer.setUp(self)

        # Instanciate the server object of the client library
        self.deejayd = DeejayDaemonAsync()
        self.deejayd.connect('localhost', self.serverPort)

        # Prepare in case we need other clients
        self.clients = [self.deejayd]

    def tearDown(self):
        for client in self.clients:
            client.disconnect()

        TestCaseWithServer.tearDown(self)

    def get_another_client(self):
        client = DeejayDaemonAsync()
        self.clients.append(client)
        return client

    def test_ping(self):
        """Ping server asynchroneously"""
        ans = self.deejayd.ping()
        self.failUnless(ans.get_contents(),
                        'Server did not respond well to ping.')

    def test_answer_callback(self):
        """Ping server asynchroneously and check for the callback to be triggered"""
        cb_called = threading.Event()
        def tcb(answer):
            cb_called.set()

        ans = self.deejayd.ping()
        ans.add_callback(tcb)
        # some seconds should be enough for the callback to be called
        cb_called.wait(4)
        self.failUnless(cb_called.isSet(), 'Answer callback was not triggered.')

    def testPlaylistSaveRetrieve(self):
        """Test playlist commands asynchroneously and callback"""

        # Get current playlist and add callback
        self.pl = None
        cb_called = threading.Event()
        def tcb(answer):
            cb_called.set()
            self.pl = answer.get_medias()

        djpl = self.deejayd.get_playlist()
        djpl.get().add_callback(tcb)

        cb_called.wait(4)
        self.failUnless(cb_called.isSet(), 'Answer callback was not triggered.')
        self.assertEqual(self.pl, [])

        # Add songs to playlist and get status
        cb_called = threading.Event()
        self.should_stop = False
        self.status = None

        def tcb_status(answer):
            cb_called.set()
            self.status = answer.get_contents()
            self.should_stop = True

        def cb_update_status(answer):
            self.deejayd.get_status().add_callback(tcb_status)

        djpl.add_path(self.test_audiodata.getRandomSongPaths(1)[0]).\
                add_callback(cb_update_status)

        while not self.should_stop:
            cb_called.wait(4)

        self.failUnless(cb_called.isSet(), 'Answer callback was not triggered.')
        self.assertEqual(self.status['playlistlength'], 1)

    def testCallbackProcess(self):
        """ Send three commands asynchroneously at the same time and check callback """

        firstcb_called = threading.Event()
        secondcb_called = threading.Event()
        thirdcb_called = threading.Event()
        def first_cb(answer):
            firstcb_called.set()

        def second_cb(answer):
            secondcb_called.set()

        def third_cb(answer):
            thirdcb_called.set()

        self.deejayd.get_audio_dir("").add_callback(first_cb)
        self.deejayd.get_playlist_list().add_callback(second_cb)
        self.deejayd.ping().add_callback(third_cb)

        firstcb_called.wait(2)
        self.failUnless(firstcb_called.isSet(), \
            '1st Answer callback was not triggered.')
        secondcb_called.wait(2)
        self.failUnless(secondcb_called.isSet(), \
            '2nd Answer callback was not triggered.')
        thirdcb_called.wait(2)
        self.failUnless(thirdcb_called.isSet(), \
            '3rd Answer callback was not triggered.')

    def test_two_clients(self):
        """Checks that it is possible to instanciate two clients in the same process."""
        client2 = self.get_another_client()
        client2.connect('localhost', self.serverPort)

        self.failUnless(self.deejayd.ping().get_contents())
        self.failUnless(client2.ping().get_contents())

    def test_subscription_another_client(self):
        """Checks that a subscription is broadcasted to another client who did not orignate the event trigger."""

        # Instanciate a second client that connects to the same server
        client2 = self.get_another_client()
        client2.connect('localhost', self.serverPort)

        self.generic_sub_bcast_test('mode', client2.set_mode, ('video', ))


# vim: ts=4 sw=4 expandtab



#!/usr/bin/env python

# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import glob,os
from distutils.command.build import build as distutils_build
from distutils.command.clean import clean as distutils_clean
from distutils.core import setup,Command
from distutils.errors import DistutilsFileError
from distutils.dep_util import newer
from distutils.dir_util import remove_tree
from distutils.spawn import find_executable
from zipfile import ZipFile

import deejayd


def force_unlink(path):
    try:
        os.unlink(path)
    except OSError:
        pass

def force_rmdir(path):
    try:
        os.rmdir(path)
    except OSError:
        pass


class build_extension(Command):
    ext_directory = None
    extension = None

    def initialize_options(self):
        pass

    def finalize_options(self):
        self.ext_directory = "extensions"
        self.extension = "deejayd-webui"
        self.ext_dir = os.path.join(self.ext_directory, self.extension)
        self.ext_path = "%s.xpi" % self.ext_dir

    def run(self):
        data_files = self.distribution.data_files

        # first remove old zip file
        self.clean()
        ext_file = ZipFile(self.ext_path, 'w')
        for root, dirs, files in os.walk(self.ext_dir):
            for f in files:
                path = os.path.join(root, f)
                ext_file.write(path, path[len(self.ext_dir):])
        ext_file.close()

        target_path = os.path.join('share', 'deejayd', 'htdocs')
        data_files.append((target_path, (self.ext_path, ), ))

    def clean(self):
        force_unlink(self.ext_path)


class build_manpages(Command):
    manpages = None
    db2mans = [
        # debian
        "/usr/share/sgml/docbook/stylesheet/xsl/nwalsh/manpages/docbook.xsl",
        # gentoo
        "/usr/share/sgml/docbook/xsl-stylesheets/manpages/docbook.xsl",
        ]
    mandir = "man/"
    executable = find_executable('xsltproc')

    def initialize_options(self):
        pass

    def finalize_options(self):
        self.manpages = glob.glob(os.path.join(self.mandir, "*.xml"))

    def __get_manpage(self, xmlmanpage):
        return xmlmanpage[:-4] # remove '.xml' at the end

    def run(self):
        data_files = self.distribution.data_files
        db2man = None
        for path in self.__class__.db2mans:
            if os.path.exists(path):
                db2man = path
                continue

        for xmlmanpage in self.manpages:
            manpage = self.__get_manpage(xmlmanpage)
            if newer(xmlmanpage, manpage):
                cmd = (self.executable, "--nonet", "-o", self.mandir, db2man,
                       xmlmanpage)
                self.spawn(cmd)

            targetpath = os.path.join("share", "man","man%s" % manpage[-1])
            data_files.append((targetpath, (manpage, ), ))

    def clean(self):
        for xmlmanpage in self.manpages:
            force_unlink(self.__get_manpage(xmlmanpage))


class build_i18n(Command):
    user_options = []
    po_package = None
    po_directory = None
    po_files = None
    executable = find_executable('msgfmt')

    def initialize_options(self):
        pass

    def finalize_options(self):
        self.po_directory = "po"
        self.po_package = "deejayd"
        self.po_files = glob.glob(os.path.join(self.po_directory, "*.po"))
        self.mo_dir = os.path.join('build', 'mo')

    def run(self):
        data_files = self.distribution.data_files

        for po_file in self.po_files:
            lang = os.path.basename(po_file[:-3])
            mo_dir =  os.path.join(self.mo_dir, lang, "LC_MESSAGES")
            mo_file = os.path.join(mo_dir, "%s.mo" % self.po_package)
            if not os.path.exists(mo_dir):
                os.makedirs(mo_dir)

            cmd = (self.executable, po_file, "-o", mo_file)
            self.spawn(cmd)

            targetpath = os.path.join("share/locale", lang, "LC_MESSAGES")
            data_files.append((targetpath, (mo_file,)))

    def clean(self):
        if os.path.isdir(self.mo_dir):
            remove_tree(self.mo_dir)


class deejayd_build(distutils_build):

    def __has_manpages(self, command):
        has_db2man = False
        for path in build_manpages.db2mans:
            if os.path.exists(path): has_db2man = True
        return self.distribution.cmdclass.has_key("build_manpages")\
            and has_db2man and build_manpages.executable != None

    def __has_i18n(self, command):
        return self.distribution.cmdclass.has_key("build_i18n")\
            and build_i18n.executable != None

    def __has_extension(self, command):
        return self.distribution.cmdclass.has_key("build_extension")

    def finalize_options(self):
        distutils_build.finalize_options(self)
        self.sub_commands.append(("build_i18n", self.__has_i18n))
        self.sub_commands.append(("build_manpages", self.__has_manpages))
        self.sub_commands.append(("build_extension", self.__has_extension))

    def clean(self):
        for subcommand_name in self.get_sub_commands():
            subcommand = self.get_finalized_command(subcommand_name)
            if hasattr(subcommand, 'clean'):
                subcommand.clean()


class deejayd_clean(distutils_clean):

    def run(self):
        distutils_clean.run(self)

        for cmd in self.distribution.command_obj.values():
            if hasattr(cmd, 'clean'):
                cmd.clean()

        force_unlink('MANIFEST')
        force_rmdir('build')


#
# data files
#
def get_data_files(walking_root, dest_dir):
    data_files = []
    for root, dirs, files in os.walk(walking_root):
        paths = [os.path.join(root, f) for f in files]
        root = root.replace(walking_root, '').strip('/')
        dest_path = os.path.join(dest_dir, root)
        data_files.append((dest_path, paths))
    return data_files

def build_data_files_list():
    data = [
        ('share/doc/deejayd', ("doc/deejayd_rpc_protocol", )),
        ('share/doc/deejayd', ("README", "NEWS", )),
        ('share/doc/deejayd', ["scripts/deejayd_rgscan"]),
        ]

    # htdocs
    data.extend(get_data_files('data/htdocs', 'share/deejayd/htdocs'))

    return data

if __name__ == "__main__":
    setup( name="deejayd", version=deejayd.__version__,
           url="http://mroy31.dyndns.org/~roy/projects/deejayd",
           description="Network controllable media player daemon",
           author="Mikael Royer, Alexandre Rossi",
           author_email="mickael.royer@gmail.com",
           license="GNU GPL v2",
           scripts=["scripts/deejayd","scripts/djc"],
           packages=["deejayd","deejayd.net","deejayd.mediadb",\
                     "deejayd.mediadb.formats",
                     "deejayd.mediadb.formats.audio",\
                     "deejayd.mediadb.formats.video","deejayd.player",\
                     "deejayd.sources","deejayd.ui","deejayd.rpc",\
                     "deejayd.database","deejayd.database.upgrade",\
                     "deejayd.database.backends","deejayd.plugins",\
                     "deejayd.webui","pytyxi"],
           package_data={'deejayd.ui': ['defaults.conf'],},
           data_files= build_data_files_list(),
           cmdclass={"build": deejayd_build,
                     "build_i18n": build_i18n,
                     "build_extension": build_extension,
                     "build_manpages": build_manpages,
                     "clean"          : deejayd_clean,
                    }
        )

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import urllib, urllib2, time, ConfigParser
import threading, Queue
try: from hashlib import md5
except ImportError: # python < 2.5
    from md5 import md5

from zope.interface import implements
from deejayd.interfaces import DeejaydError
from deejayd.ui import log
from deejayd.plugins import IPlayerPlugin, PluginError
from deejayd.utils import str_encode

class AudioScrobblerFatalError(DeejaydError): pass
class AudioScrobblerError(DeejaydError):

    def __init__(self, msg):
        self.message = _("AudioScrobbler Error: %s") % str_encode(msg)


class AudioScrobblerPlugin:
    implements(IPlayerPlugin)
    NAME="audioscrobbler"

    AUDIOSCROBBLER_URL = "post.audioscrobbler.com"
    AUDIOSCROBBLER_PROTOCOL_VERSION = "1.2.1"
    AUDIOSCROBBLER_CLIENT="tst"
    AUDIOSCROBBLER_VERSION="1.0"
    AUDIOSCROBBLER_WAIT_INTERVAL=30

    def __init__(self, config):
        self.enabled = True
        self.authenticated = False
        self.last_request = None
        self.session_id = None
        self.nowplaying_url = None
        self.submit_url = None
        self.should_stop = threading.Event()

        self.queue = Queue.Queue()
        try:
            self.__auth_details = {
                    "login": config.get("lastfm", "login"),
                    "password": config.get("lastfm", "password"),
                }
        except (ConfigParser.NoSectionError, ConfigParser.NoOptionError):
            raise PluginError(_("Lastfm configuration has not been set."))
        # start thread
        self.thread = threading.Thread(target=self.run)
        self.thread.setDaemon(True)     # exit if only this thread is left
        self.thread.start()

    def on_media_played(self, media):
        if self.enabled and media["type"] == "song":
            self.queue.put((media, self.__get_stamp(media)))

    def close(self):
        self.should_stop.set()

    def __open_request(self, request):
        try:
            url_handle = urllib2.urlopen(request)
        except urllib2.HTTPError, error:
            err_msg = _("Unable to connect to server: %s - %s")
            code = str_encode(error.code, errors="ignore")
            msg = str_encode(error.msg, errors="ignore")
            raise AudioScrobblerError(err_msg % (code, msg))
        except urllib2.URLError, error:
            args = getattr(error.reason, 'args', None)
            code = '000'
            message = str(error)
            if args is not None:
                if len(args) == 1:
                    message = error.reason.args[0]
                elif len(args) == 2:
                    code = str(error.reason.args[0])
                    message = error.reason.args[1]
            code = str_encode(code, errors="ignore")
            message = str_encode(message, errors="ignore")
            err_msg = _("Unable to connect to server: %s - %s")
            raise AudioScrobblerError(err_msg % (code, message))

        self.last_request = int(time.time())
        return url_handle

    def auth(self):
        if self.authenticated:
            return True

        timestamp = int(time.time())
        token = md5(self.__auth_details["password"]+str(timestamp)).hexdigest()
        p = {
            "hs": "true",
            "u": self.__auth_details["login"],
            "c": self.AUDIOSCROBBLER_CLIENT,
            "v": self.AUDIOSCROBBLER_VERSION,
            "p": self.AUDIOSCROBBLER_PROTOCOL_VERSION,
            "t": str(timestamp),
            "a": token,
        }
        plist = [(k, urllib.quote_plus(v.encode('utf8'))) for k, v in p.items()]

        authparams = urllib.urlencode(plist)
        url = 'http://%s/?%s' % (self.AUDIOSCROBBLER_URL, authparams)
        req = urllib2.Request(url)
        url_handle = self.__open_request(req)

        # check answer
        response = url_handle.read().rstrip().split("\n")
        if len(response) == 0:
            raise AudioScrobblerError(_('Got nothing back from the server'))

        status = response.pop(0)
        if status == "OK":
            self.session_id = response.pop(0)
            self.nowplaying_url = response.pop(0)
            self.submit_url = response.pop(0)
            self.authenticated = True
        elif status == "UPTODATE":
            self.authenticated = True
        elif status == "BADAUTH":
            raise AudioScrobblerFatalError(_("Bad login/password"))
        elif status == "BADTIME":
            raise AudioScrobblerFatalError(_("Bad time"))
        elif status == "BANNED":
            raise AudioScrobblerFatalError(_("Application has be banned"))
        elif status.startswith("FAILED"):
            raise AudioScrobblerError(_("Failed to handshake: %s") % status)
        else:
            raise AudioScrobblerFatalError(_("Unknown status %s") % status)

    def submit_track(self, queue):
        data = {"s": self.session_id}
        for i, (track, stamp) in enumerate(queue):
            data["a[%d]" % i] = track["artist"].encode('utf-8')
            data["t[%d]" % i] = track["title"].encode('utf-8')
            data["l[%d]" % i] = track["length"].encode('utf-8')
            data["b[%d]" % i] = track['album'].encode('utf-8')
            data["m[%d]" % i] = "".encode('utf-8')
            data["i[%d]" % i] = stamp
            data["o[%d]" % i] = "P".encode('utf-8')
            data["n[%d]" % i] = "".encode('utf-8')
            data["r[%d]" % i] = "".encode('utf-8')

        postdata = urllib.urlencode(data)
        req = urllib2.Request(url=self.submit_url, data=postdata)
        url_handle = self.__open_request(req)

        response = url_handle.read().rstrip().split("\n")
        if len(response) == 0:
            raise AudioScrobblerError(_('Got nothing back from the server'))
        status = response.pop(0)
        if status == "OK":
            return True
        if status == "BADSESSION":
            raise AudioScrobblerFatalError(_("Bad session ID"))
        elif status.startswith("FAILED"):
            raise AudioScrobblerError(_("Failed to submit songs: %s") % status)
        else:
            raise AudioScrobblerError(_("Unknown status %s") % status)

    def run(self):
        while not self.should_stop.isSet():
            try: self.auth()
            except AudioScrobblerFatalError, ex: # fatal error, disable plugin
                log.err(_("Fatal error in audioscrobbler: %s") % ex)
                log.err(_("Disable audioscrobbler plugin"))
                self.enabled = False
                break
            except AudioScrobblerError, ex: # log error and try later
                log.err(str(ex))
            else:
                tracks = []
                if not self.queue.empty():
                    if (int(time.time()) - self.last_request) > 3600:
                        log.info(_("Last FM : force reauthentification"))
                        self.authenticated = False
                        continue
                    for i in range(1,10):
                        if self.queue.empty():
                            break
                        tracks.append(self.queue.get())

                if len(tracks):
                    submission_failures = 0
                    while submission_failures < 3:
                        try: self.submit_track(tracks)
                        except AudioScrobblerError, err:
                            log.err(_("Unable to submit songs: %s") % err)
                            # wait 10 sec before retry
                            self.should_stop.wait(10)
                            submission_failures += 1
                        except AudioScrobblerFatalError, err:
                            log.err(_("Fatal error in submission process: %s")\
                                     % err)
                            log.err(_("Force lastfm reauthentification"))
                            submission_failures = 3
                            break
                        else:
                            break
                    if submission_failures == 3:
                        self.authenticated = False
                        for track in tracks:
                            self.queue.put(track)
            self.should_stop.wait(self.AUDIOSCROBBLER_WAIT_INTERVAL)

    def __get_stamp(self, track):
        stamp = str(int(time.time())--int(int(track['length'])/2))
        return stamp.encode("utf-8")

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""The Deejayd python client library"""

import socket, asyncore, threading
import sys,time

from Queue import Queue, Empty

import deejayd.interfaces
from deejayd.interfaces import DeejaydError, DeejaydKeyValue, DeejaydSignal
from deejayd.rpc import Fault, DEEJAYD_PROTOCOL_VERSION
from deejayd.rpc.jsonbuilders import JSONRPCRequest, Get_json_filter
from deejayd.rpc.jsonparsers import loads_response, Parse_json_filter


MSG_DELIMITER = 'ENDJSON\n'
MAX_BANNER_LENGTH = 50


class DeejaydAnswer(deejayd.interfaces.DeejaydAnswer):

    def __init__(self, server = None):
        deejayd.interfaces.DeejaydAnswer.__init__(self)
        self.answer_received = threading.Event()
        self.callbacks = []
        self.server = server
        self.id = None

    def set_id(self, id):
        self.id = id

    def get_id(self):
        return self.id

    def wait(self):
        self.answer_received.wait()

    def _received(self, contents):
        self.contents = contents
        self.answer_received.set()
        self._run_callbacks()

    def set_error(self, msg):
        deejayd.interfaces.DeejaydAnswer.set_error(self, msg)
        self.answer_received.set()
        self._run_callbacks()

    def get_contents(self):
        self.wait()
        return deejayd.interfaces.DeejaydAnswer.get_contents(self)

    def add_callback(self, cb):
        if self.answer_received.isSet():
            cb(self)
        else:
            self.callbacks.append(cb)

    def _run_callbacks(self):
        self.wait()
        for cb in self.callbacks:
            cb(self)


class DeejaydKeyValue(DeejaydAnswer, deejayd.interfaces.DeejaydKeyValue):

    def __init__(self, server=None):
        DeejaydAnswer.__init__(self, server)


class DeejaydList(DeejaydAnswer, deejayd.interfaces.DeejaydList):

    def __init__(self, server=None):
        DeejaydAnswer.__init__(self, server)


class DeejaydFileList(DeejaydAnswer, deejayd.interfaces.DeejaydFileList):

    def __init__(self, server = None):
        deejayd.interfaces.DeejaydFileList.__init__(self)
        DeejaydAnswer.__init__(self, server)


class DeejaydMediaList(DeejaydAnswer, deejayd.interfaces.DeejaydMediaList):

    def __init__(self, server = None):
        deejayd.interfaces.DeejaydMediaList.__init__(self)
        DeejaydAnswer.__init__(self, server)


class DeejaydDvdInfo(DeejaydAnswer, deejayd.interfaces.DeejaydDvdInfo):

    def __init__(self, server = None):
        deejayd.interfaces.DeejaydDvdInfo.__init__(self)
        DeejaydAnswer.__init__(self, server)


class DeejaydStaticPlaylist(deejayd.interfaces.DeejaydStaticPlaylist):

    def __init__(self, server, pl_id, name):
        self.server = server
        self.__pl_id = pl_id
        self.__name = name

    def get(self, first=0, length=-1):
        params = [self.__pl_id, first]
        if length != -1:
            params.append(length)
        cmd = JSONRPCRequest('recpls.get', params)
        ans = DeejaydMediaList(self.server)
        return self.server._send_command(cmd, ans)

    def __add(self, values, type):
        cmd = JSONRPCRequest('recpls.staticAdd', [self.__pl_id, values, type])
        return self.server._send_command(cmd)

    def add_songs(self, song_ids):
        return self.__add(song_ids, "id")

    def add_paths(self, paths):
        return self.__add(paths, "path")


class DeejaydMagicPlaylist(deejayd.interfaces.DeejaydMagicPlaylist):

    def __init__(self, server, pl_id, name):
        self.server = server
        self.__pl_id = pl_id
        self.__name = name

    def get(self, first=0, length=-1):
        params = [self.__pl_id, first]
        if length != -1:
            params.append(length)
        cmd = JSONRPCRequest('recpls.get', params)
        ans = DeejaydMediaList(self.server)
        return self.server._send_command(cmd, ans)

    def add_filter(self, filter):
        jfilter = Get_json_filter(filter).dump()
        cmd = JSONRPCRequest('recpls.magicAddFilter', [self.__pl_id, jfilter])
        return self.server._send_command(cmd)

    def remove_filter(self, filter):
        jfilter = Get_json_filter(filter).dump()
        cmd = JSONRPCRequest('recpls.magicRemoveFilter',\
                [self.__pl_id, jfilter])
        return self.server._send_command(cmd)

    def clear_filters(self):
        cmd = JSONRPCRequest('recpls.magicClearFilter', [self.__pl_id])
        return self.server._send_command(cmd)

    def get_properties(self):
        cmd = JSONRPCRequest('recpls.magicGetProperties', [self.__pl_id])
        return self.server._send_command(cmd, DeejaydKeyValue())

    def set_property(self, key, value):
        cmd = JSONRPCRequest('recpls.magicSetProperty',\
                [self.__pl_id, key, value])
        return self.server._send_command(cmd)


class DeejaydWebradioList(deejayd.interfaces.DeejaydWebradioList):

    def __init__(self, server):
        self.server = server

    def get(self, first = 0, length = None):
        params = [first]
        if length != None:
            params.append(length)
        cmd = JSONRPCRequest('webradio.get', params)
        ans = DeejaydMediaList(self)
        return self.server._send_command(cmd, ans)

    def get_available_sources(self):
        cmd = JSONRPCRequest('webradio.getAvailableSources', [])
        return self.server._send_command(cmd, DeejaydKeyValue())

    def get_source_categories(self, source_name):
        cmd = JSONRPCRequest('webradio.getSourceCategories', [source_name])
        return self.server._send_command(cmd, DeejaydList())

    def set_source(self, source_name):
        cmd = JSONRPCRequest('webradio.setSource', [source_name])
        return self.server._send_command(cmd)

    def set_source_categorie(self, categorie):
        cmd = JSONRPCRequest('webradio.setSourceCategorie', [categorie])
        return self.server._send_command(cmd)

    def add_webradio(self, name, url):
        cmd = JSONRPCRequest('webradio.localAdd', [name, url])
        return self.server._send_command(cmd)

    def delete_webradios(self, wr_ids):
        cmd = JSONRPCRequest('webradio.localRemove', [wr_ids])
        return self.server._send_command(cmd)

    def clear(self):
        cmd = JSONRPCRequest('webradio.localClear', [])
        return self.server._send_command(cmd)


class DeejaydQueue(deejayd.interfaces.DeejaydQueue):

    def __init__(self, server):
        self.server = server

    def get(self, first = 0, length = None):
        params = [first]
        if length != None:
            params.append(length)
        cmd = JSONRPCRequest('queue.get', params)
        ans = DeejaydMediaList(self)
        return self.server._send_command(cmd, ans)

    def add_songs(self, song_ids, pos = None):
        params = [song_ids]
        if pos != None:
            params.append(pos)
        cmd = JSONRPCRequest('queue.addIds', params)
        return self.server._send_command(cmd)

    def add_paths(self, paths, pos = None):
        params = [paths]
        if pos != None:
            params.append(pos)
        cmd = JSONRPCRequest('queue.addPath', params)
        return self.server._send_command(cmd)

    def load_playlists(self, pl_ids, pos = None):
        params = [pl_ids]
        if pos != None:
            params.append(pos)
        cmd = JSONRPCRequest('queue.loads', params)
        return self.server._send_command(cmd)

    def clear(self):
        cmd = JSONRPCRequest('queue.clear', [])
        return self.server._send_command(cmd)

    def move(self, ids, new_pos):
        cmd = JSONRPCRequest('queue.move', [ids, new_pos])
        return self.server._send_command(cmd)

    def del_songs(self, ids):
        cmd = JSONRPCRequest('queue.remove', [ids])
        return self.server._send_command(cmd)


class DeejaydPlaylistMode(deejayd.interfaces.DeejaydPlaylistMode):

    def __init__(self, server):
        self.server = server

    def get(self, first = 0, length = None):
        params = [first]
        if length != None:
            params.append(length)
        cmd = JSONRPCRequest('playlist.get', params)
        ans = DeejaydMediaList(self)
        return self.server._send_command(cmd, ans)

    def save(self, name):
        cmd = JSONRPCRequest('playlist.save', [name])
        return self.server._send_command(cmd, DeejaydKeyValue())

    def add_songs(self, song_ids, pos = None):
        params = [song_ids]
        if pos != None:
            params.append(pos)
        cmd = JSONRPCRequest('playlist.addIds', params)
        return self.server._send_command(cmd)

    def add_paths(self, paths, pos = None):
        params = [paths]
        if pos != None:
            params.append(pos)
        cmd = JSONRPCRequest('playlist.addPath', params)
        return self.server._send_command(cmd)

    def loads(self, pl_ids, pos = None):
        params = [pl_ids]
        if pos != None:
            params.append(pos)
        cmd = JSONRPCRequest('playlist.loads', params)
        return self.server._send_command(cmd)

    def move(self, ids, new_pos):
        cmd = JSONRPCRequest('playlist.move', [ids, new_pos])
        return self.server._send_command(cmd)

    def shuffle(self):
        cmd = JSONRPCRequest('playlist.shuffle', [])
        return self.server._send_command(cmd)

    def clear(self):
        cmd = JSONRPCRequest('playlist.clear', [])
        return self.server._send_command(cmd)

    def del_songs(self, ids):
        cmd = JSONRPCRequest('playlist.remove', [ids])
        return self.server._send_command(cmd)


class DeejaydPanel(deejayd.interfaces.DeejaydPanel):

    def __init__(self, server):
        self.server = server

    def get(self, first = 0, length = None):
        params = [first]
        if length != None:
            params.append(length)
        cmd = JSONRPCRequest('panel.get', params)
        ans = DeejaydMediaList(self)
        return self.server._send_command(cmd, ans)

    def get_panel_tags(self):
        cmd = JSONRPCRequest('panel.tags', [])
        return self.server._send_command(cmd, DeejaydList())

    def get_active_list(self):
        cmd = JSONRPCRequest('panel.activeList', [])
        return self.server._send_command(cmd, DeejaydKeyValue())

    def set_active_list(self, type, pl_id=""):
        cmd = JSONRPCRequest('panel.setActiveList', [type, pl_id])
        return self.server._send_command(cmd)

    def set_panel_filters(self, tag, values):
        if not isinstance(values, list):
            values = [values]
        cmd = JSONRPCRequest('panel.setFilter', [tag, values])
        return self.server._send_command(cmd)

    def remove_panel_filters(self, tag):
        cmd = JSONRPCRequest('panel.removeFilter', [tag])
        return self.server._send_command(cmd)

    def clear_panel_filters(self):
        cmd = JSONRPCRequest('panel.clearFilter', [])
        return self.server._send_command(cmd)

    def set_search_filter(self, tag, value):
        cmd = JSONRPCRequest('panel.setSearch', [tag, value])
        return self.server._send_command(cmd)

    def clear_search_filter(self):
        cmd = JSONRPCRequest('panel.clearSearch', [])
        return self.server._send_command(cmd)

    def set_sorts(self, sort):
        cmd = JSONRPCRequest('panel.setSort', [sort])
        return self.server._send_command(cmd)


class DeejaydVideo(deejayd.interfaces.DeejaydVideo):

    def __init__(self, server):
        self.server = server

    def get(self, first = 0, length = None):
        params = [first]
        if length != None:
            params.append(length)
        cmd = JSONRPCRequest('video.get', params)
        ans = DeejaydMediaList(self)
        return self.server._send_command(cmd, ans)

    def set(self, value, type = "directory"):
        cmd = JSONRPCRequest('video.set', [value, type])
        return self.server._send_command(cmd)

    def set_sorts(self, sort):
        cmd = JSONRPCRequest('video.sort', [sort])
        return self.server._send_command(cmd)


class ConnectError(deejayd.interfaces.DeejaydError):
    pass


class _DeejayDaemon(deejayd.interfaces.DeejaydCore):
    """Abstract class for a deejay daemon client."""

    def __init__(self):
        deejayd.interfaces.DeejaydCore.__init__(self)

        self.expected_answers_queue = Queue()
        self.connected = False

    def connect(self, host, port, ignore_version=False):
        raise NotImplementedError

    def is_connected(self):
        return self.connected

    def disconnect(self):
        raise NotImplementedError

    def _version_from_banner(self, banner_line):
        tokenized_banner = banner_line.split(' ')
        try:
            version = tokenized_banner[2]
        except IndexError:
            raise ValueError
        else:
            numerical_version = map(int, version.split('.'))
            try:
                if tokenized_banner[3] == 'protocol':
                    protocol_version = tokenized_banner[4]
                else:
                    raise IndexError
            except IndexError:
                # Assume protocol is first one
                protocol_version = 1
            else:
                protocol_version = int(protocol_version)
            return numerical_version, protocol_version

    def _version_is_supported(self, versions):
        numerical_version = versions[0]
        protocol_version = versions[1]
        return protocol_version == DEEJAYD_PROTOCOL_VERSION

    def _send_simple_command(self, cmd_name):
        cmd = JSONRPCRequest(cmd_name, [])
        return self._send_command(cmd)

    def get_recorded_playlist(self, pl_id):
        return DeejaydStaticPlaylist(self, pl_id)

    def get_playlist(self):
        return DeejaydPlaylistMode(self)

    def get_panel(self):
        return DeejaydPanel(self)

    def get_webradios(self):
        return DeejaydWebradioList(self)

    def get_video(self):
        return DeejaydVideo(self)

    def get_queue(self):
        return DeejaydQueue(self)

    def ping(self):
        return self._send_simple_command('ping')

    def set_mode(self, mode):
        cmd = JSONRPCRequest('setmode', [mode])
        return self._send_command(cmd)

    def get_mode(self):
        cmd = JSONRPCRequest('availablemodes', [])
        return self._send_command(cmd, DeejaydKeyValue())

    def set_option(self, source, option_name, option_value):
        cmd = JSONRPCRequest('setOption', [source, option_name, option_value])
        return self._send_command(cmd)

    def get_status(self):
        cmd = JSONRPCRequest('status', [])
        return self._send_command(cmd, DeejaydKeyValue())

    def get_stats(self):
        cmd = JSONRPCRequest('stats', [])
        return self._send_command(cmd, DeejaydKeyValue())

    def play_toggle(self):
        return self._send_simple_command('player.playToggle')

    def stop(self):
        return self._send_simple_command('player.stop')

    def previous(self):
        return self._send_simple_command('player.previous')

    def next(self):
        return self._send_simple_command('player.next')

    def seek(self, pos, relative = False):
        cmd = JSONRPCRequest('player.seek', [pos, relative])
        return self._send_command(cmd)

    def get_current(self):
        cmd = JSONRPCRequest('player.current', [])
        return self._send_command(cmd,DeejaydMediaList())

    def go_to(self, id, id_type = "id", source = None):
        params = [id, id_type]
        if source:
            params.append(source)
        cmd = JSONRPCRequest('player.goto', params)
        return self._send_command(cmd)

    def set_volume(self, volume):
        cmd = JSONRPCRequest('player.setVolume', [volume])
        return self._send_command(cmd)

    def set_player_option(self, name, value):
        cmd = JSONRPCRequest('player.setPlayerOption', [name, value])
        return self._send_command(cmd)

    def update_audio_library(self, force = False):
        cmd = JSONRPCRequest('audiolib.update', [force])
        return self._send_command(cmd, DeejaydKeyValue())

    def update_video_library(self, force = False):
        cmd = JSONRPCRequest('videolib.update', [force])
        return self._send_command(cmd, DeejaydKeyValue())

    def get_audio_dir(self, dir = ""):
        ans = DeejaydFileList(self)
        cmd = JSONRPCRequest('audiolib.getDir', [dir])
        return self._send_command(cmd, ans)

    def audio_search(self, pattern, type = 'all'):
        ans = DeejaydMediaList(self)
        cmd = JSONRPCRequest('audiolib.search', [pattern, type])
        return self._send_command(cmd, ans)

    def mediadb_list(self, tag, filter):
        params = [tag]
        if filter is not None:
            params.append(Get_json_filter(filter).dump())
        ans = DeejaydList(self)
        cmd = JSONRPCRequest('audiolib.taglist', params)
        return self._send_command(cmd, ans)

    def get_video_dir(self, dir = ""):
        ans = DeejaydFileList(self)
        cmd = JSONRPCRequest('videolib.getDir', [dir])
        return self._send_command(cmd, ans)

    def create_recorded_playlist(self, name, type):
        cmd = JSONRPCRequest('recpls.create', [name, type])
        return self._send_command(cmd, DeejaydKeyValue())

    def get_recorded_playlist(self, pl_id, name, type):
        if type == "static":
            return DeejaydStaticPlaylist(self, pl_id, name)
        elif type == "magic":
            return DeejaydMagicPlaylist(self, pl_id, name)

    def erase_playlist(self, pl_ids):
        cmd = JSONRPCRequest('recpls.erase', [pl_ids])
        return self._send_command(cmd)

    def get_playlist_list(self):
        cmd = JSONRPCRequest('recpls.list', [])
        return self._send_command(cmd,DeejaydMediaList())

    def set_media_rating(self, media_ids, rating, type = "audio"):
        cmd = JSONRPCRequest('setRating', [media_ids, rating, type])
        return self._send_command(cmd)

    def dvd_reload(self):
        return self._send_simple_command('dvd.reload')

    def get_dvd_content(self):
        cmd = JSONRPCRequest('dvd.get', [])
        ans = DeejaydDvdInfo(self)
        return self._send_command(cmd, ans)

    def _build_answer(self, msg):
        try: msg = loads_response(msg)
        except Fault, f:
            raise DeejaydError("JSONRPC error - %s - %s" % (f.code, f.message))
        else:
            if msg["id"] is None: # it is a notification
                result =  msg["result"]["answer"]
                type = msg["result"]["type"]
                if type == 'signal':
                    signal = DeejaydSignal(result["name"], result["attrs"])
                    return self._dispatch_signal(signal)
            else:
                expected_answer = self.expected_answers_queue.get()
                if expected_answer.id != msg["id"]:
                    raise DeejaydError("Bad id for JSON server answer")

                if msg["error"] is not None: # an error is returned
                    expected_answer.set_error("Deejayd Server Error - %s - %s"\
                            % (msg["error"]["code"], msg["error"]["message"]))
                else:
                    result =  msg["result"]["answer"]
                    type = msg["result"]["type"]
                    if type == 'fileAndDirList':
                        expected_answer.set_rootdir(result["root"])
                        expected_answer.set_files(result["files"])
                        expected_answer.set_directories(result["directories"])
                    elif type == 'mediaList':
                        expected_answer.set_medias(result["medias"])
                        expected_answer.set_media_type(result["media_type"])
                        if "filter" in result and result["filter"] is not None:
                            expected_answer.set_filter(\
                                    Parse_json_filter(result["filter"]))
                        if "sort" in result:
                            expected_answer.set_sort(result["sort"])
                    elif type == 'dvdInfo':
                        expected_answer.set_dvd_content(result)
                    expected_answer._received(result)


class DeejayDaemonSync(_DeejayDaemon):
    """Synchroneous deejayd client library."""

    def __init__(self):
        _DeejayDaemon.__init__(self)

        self.socket_to_server = socket.socket(socket.AF_INET,
                                              socket.SOCK_STREAM)
        self.__timeout = 2
        self.socket_to_server.settimeout(self.__timeout)
        self.next_msg = ""

    def connect(self, host, port, ignore_version=False):
        if self.connected:
            self.disconnect()

        self.host = host
        self.port = port

        try: self.socket_to_server.connect((self.host, self.port))
        except socket.timeout, msg:
            # reset connection
            self._reset_socket()
            raise ConnectError('Connection timeout')
        except socket.error, msg:
            raise ConnectError('Connection with server failed : %s' % msg)

        self.socket_to_server.settimeout(None)
        socketFile = self.socket_to_server.makefile()

        # Catch version
        banner_line = socketFile.readline()
        self.connected = True
        if not banner_line.startswith("OK DEEJAYD")\
        or len(banner_line) > MAX_BANNER_LENGTH:
            self.disconnect()
            raise ConnectError('Connection with server failed')
        try:
            versions = self._version_from_banner(banner_line)
        except ValueError:
            self.disconnect()
            raise ConnectError('Initial version dialog with server failed')
        if not ignore_version and not self._version_is_supported(versions):
            self.disconnect()
            raise ConnectError('This server version protocol is not handled by this client version')

    def _send_command(self, cmd, expected_answer = None):
        # Set a default answer by default
        if expected_answer == None:
            expected_answer = DeejaydAnswer(self)
        expected_answer.set_id(cmd.get_id())
        self.expected_answers_queue.put(expected_answer)

        self._sendmsg(cmd.to_json())

        rawmsg = self._readmsg()
        self._build_answer(rawmsg)
        return expected_answer

    def disconnect(self):
        if not self.connected:
            return

        self.socket_to_server.settimeout(self.__timeout)
        try: self._send_simple_command('close').get_contents()
        except socket.timeout: pass

        self._reset_socket()
        self.connected = False
        self.host = None
        self.port = None

    def _reset_socket(self):
        self.socket_to_server.close()
        # New Socket setup
        self.socket_to_server = socket.socket(socket.AF_INET,
                                              socket.SOCK_STREAM)
        self.socket_to_server.settimeout(self.__timeout)

    def _sendmsg(self, buf):
        self.socket_to_server.send(buf + MSG_DELIMITER)

    def _readmsg(self):
        msg_chunks = ''
        msg_chunk = ''
        def split_msg(msg,index):
            return (msg[0:index], msg[index+len(MSG_DELIMITER):len(msg)])

        while 1:
            try: index = self.next_msg.index(MSG_DELIMITER)
            except ValueError: pass
            else:
                (rs,self.next_msg) = split_msg(self.next_msg, index)
                break

            msg_chunk = self.socket_to_server.recv(4096)
            # socket.recv returns an empty string if the socket is closed, so
            # catch this.
            if msg_chunk == '':
                raise socket.error()

            msg_chunks += msg_chunk
            try: index = msg_chunks.index(MSG_DELIMITER)
            except ValueError: pass
            else:
                (rs,self.next_msg) = split_msg(msg_chunks, index)
                break

        return rs

    # No subscription for the sync client
    def subscribe(self, signal_name, callback): raise NotImplementedError
    def unsubscribe(self, sub_id): raise NotImplementedError


class _DeejaydSocket(asyncore.dispatcher):
    ac_in_buffer_size = 256
    ac_out_buffer_size = 256

    def __init__(self, socket_map, deejayd, ignore_version=False):
        asyncore.dispatcher.__init__(self, map=socket_map)
        self.deejayd = deejayd
        self.ignore_version = ignore_version

        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)

        self.socket_die_callback = []
        self.__connect_callback = []

        self.state = "disconnected"
        self.msg_chunks = ""
        self.next_msg = ""

    def add_connect_callback(self, callback):
        self.__connect_callback.append(callback)

    def add_socket_die_callback(self, callback):
        self.socket_die_callback.append(callback)

    def __error_callbacks(self, msg):
        for cb in self.socket_die_callback:
            cb(msg)

    def handle_error(self):
        t, v, tb = sys.exc_info()
        assert tb # Must have a traceback
        if self.state != "json_protocol":
            # error appears when we try to connect
            for cb in self.__connect_callback:
                cb(False,str(v))

    def handle_connect(self):
        self.state = "connected"

    def handle_close(self):
        if self.state == "json_protocol":
            self.__error_callbacks('disconnected')
        self.state = "disconnected"
        self.close()

    def handle_read(self):
        def split_msg(msg,index):
            return (msg[0:index], msg[index+len(MSG_DELIMITER):len(msg)])

        if self.state == "connected":
            # Catch banner until first newline char
            banner_line = ''
            newchar = ''
            while newchar != '\n' and len(banner_line) < MAX_BANNER_LENGTH:
                banner_line += newchar
                newchar = self.recv(1)
            if banner_line.startswith("OK DEEJAYD"):
                try:
                    versions = self.deejayd._version_from_banner(banner_line)
                except ValueError:
                    for cb in self.__connect_callback:
                        cb(False, 'Initial version dialog with server failed')
                if self.ignore_version\
                or self.deejayd._version_is_supported(versions):
                    self.state = 'json_protocol'
                    # now we are sure to be connected
                    for cb in self.__connect_callback:
                        cb(True,"")
                else:
                    for cb in self.__connect_callback:
                        cb(False,\
          'This server version protocol is not handled by this client version')

        elif self.state == "json_protocol":
            msg_chunk = self.recv(256)
            # socket.recv returns an empty string if the socket is closed, so
            # catch this.
            if msg_chunk == '':
                self.close()
            self.msg_chunks += msg_chunk
            try: index = self.msg_chunks.index(MSG_DELIMITER)
            except ValueError: return
            else:
                (rawmsg,self.next_msg) = split_msg(self.msg_chunks, index)
                self.msg_chunks = ""
                self.answer_received(rawmsg)

                while self.next_msg != '':
                    try: index = self.next_msg.index(MSG_DELIMITER)
                    except ValueError:
                        self.msg_chunks = self.next_msg
                        self.next_msg = ""
                        break
                    else:
                        (rawmsg,self.next_msg) = split_msg(self.next_msg, index)
                        self.answer_received(rawmsg)

        elif self.state == "disconnected":
            # This should not happen
            raise AttributeError

    def answer_received(self,rawmsg):
        try: self.deejayd._build_answer(rawmsg)
        except DeejaydError:
            self.__error_callbacks("Unable to parse server answer : %s" %rawmsg)
            self.close()

    def handle_write(self):
        cmd = self.deejayd.command_queue.get()
        self.send(cmd.to_json()+MSG_DELIMITER)

    def writable(self):
        if self.state != "json_protocol": return False
        return not self.deejayd.command_queue.empty()


class _DeejaydSocketThread:

    def __init__(self):
        self.__th = None
        self.socket_map = {}

    # Thanks
    # http://mail.python.org/pipermail/python-list/2004-November/290798.html
    def __asyncore_loop(self):
        self.__stop_asyncore_loop = False
        while self.socket_map and not self.__stop_asyncore_loop:
            asyncore.loop(timeout=1, map=self.socket_map, count=1)

    def stop(self):
        self.__stop_asyncore_loop = True
        self.__th.join()

    def start(self):
        self.__th = threading.Thread(target=self.__asyncore_loop)
        self.__th.start()


class DeejayDaemonAsync(_DeejayDaemon):
    """Completely aynchroneous deejayd client library."""

    # There is only one socket thread for all the async client instances
    socket_thread = _DeejaydSocketThread()

    def __init__(self):
        _DeejayDaemon.__init__(self)
        #socket.setdefaulttimeout(5)
        self.command_queue = Queue()
        self.__con_cb = []
        self.__err_cb = []
        self.socket_to_server = None

    def __create_socket(self, ignore_version=False):
        self.socket_to_server = _DeejaydSocket(self.socket_thread.socket_map,
                                               self, ignore_version)
        self.socket_to_server.add_socket_die_callback(\
            self.__make_answers_obsolete)
        self.socket_to_server.add_socket_die_callback(self.__disconnect)
        self.socket_to_server.add_connect_callback(self.__connect_callback)

    def __disconnect(self, msg):
        # socket die, so thread do not need to be stopped
        # just reset the socket
        self.connected = False
        del self.socket_to_server
        self.socket_to_server = None
        # execute error callback
        for cb in self.__err_cb:
            cb(msg)

    def __make_answers_obsolete(self, msg):
        msg = "Could not obtain answer from server : " + msg
        try:
            while True:
                self.expected_answers_queue.get_nowait().set_error(msg)
        except Empty:
            # There is no more answer there, so no need to set any more errors.
            pass

    def __connect_callback(self, rs, msg = ""):
        self.connected = rs
        if not rs: # error happens, reset socket
            del self.socket_to_server
            self.socket_to_server = None
        for cb in self.__con_cb:
            cb(rs, msg)

    def add_connect_callback(self,cb):
        self.__con_cb.append(cb)

    def add_error_callback(self,cb):
        self.__err_cb.append(cb)

    def connect(self, host, port, ignore_version=False):
        if self.connected:
            self.disconnect()

        self.__create_socket(ignore_version)
        try: self.socket_to_server.connect((host, port))
        except socket.connecterror, msg:
            raise ConnectError('Connection error %s' % str(msg))

        if len(self.socket_thread.socket_map) < 2:
            # This is the first client to ask for a connection
            self.socket_thread.start()

    def disconnect(self):
        if self.socket_to_server != None:
            self.__stop_thread()
        self.connected = False
        # clear subscriptions
        self._clear_subscriptions()

    def __stop_thread(self):
        # terminate socket thread
        self.socket_to_server.close()
        if len(self.socket_thread.socket_map) < 1:
            # This is the last client to disconnect
            self.socket_thread.stop()
        del self.socket_to_server
        self.socket_to_server = None

    def _send_command(self, cmd, expected_answer = None):
        # Set a default answer by default
        if expected_answer == None:
            expected_answer = DeejaydAnswer(self)
        expected_answer.set_id(cmd.get_id())

        self.expected_answers_queue.put(expected_answer)
        self.command_queue.put(cmd)
        return expected_answer

    def subscribe(self, signal_name, callback):
        if signal_name not in self.get_subscriptions().values():
            cmd = JSONRPCRequest('signal.setSubscription', [signal_name, True])
            ans = self._send_command(cmd)
            # Subscription are sync because there should be a way to tell the
            # client that subscription failed.
            ans.get_contents()
        return _DeejayDaemon.subscribe(self, signal_name, callback)

    def unsubscribe(self, sub_id):
        try:
            signal_name = self.get_subscriptions()[sub_id]
        except KeyError:
            # DeejaydError will be raised at local unsubscription
            pass
        _DeejayDaemon.unsubscribe(self, sub_id)
        # Remote unsubscribe if last local subscription laid off
        if signal_name not in self.get_subscriptions().values():
            cmd = JSONRPCRequest('signal.setSubscription', [signal_name, False])
            ans = self._send_command(cmd)
            # As subscription, unsubscription is sync.
            ans.get_contents()


#
# HTTP client
#
import httplib

class DeejayDaemonHTTP(_DeejayDaemon):
    """HTTP deejayd client library."""

    def __init__(self, host, port = 6880):
        _DeejayDaemon.__init__(self)
        self.host = host
        self.port = port
        self.connection = httplib.HTTPConnection(self.host, self.port)
        self.hdrs = {
                "Content-Type": "text/json",
                "Accept": "text/json",
                "User-Agent": "Deejayd Client Library",
            }

    def test_compatibility(self):
        # get server informations
        cmd = JSONRPCRequest('serverInfo', [])
        ans = self._send_command(cmd, DeejaydKeyValue())

        versions = (ans["server_version"], ans["protocol_version"])
        return self._version_is_supported(versions)

    def _send_command(self, cmd, expected_answer = None):
        # Set a default answer by default
        if expected_answer == None:
            expected_answer = DeejaydAnswer(self)
        expected_answer.set_id(cmd.get_id())
        self.expected_answers_queue.put(expected_answer)

        # send http request
        try: self.connection.request("POST", "/rpc/", cmd.to_json(), self.hdrs)
        except Exception, ex:
            raise DeejaydError("Unable to send request : %s" % str(ex))

        # get answer
        response = self.connection.getresponse()
        if response.status != 200:
            raise DeejaydError("Server return error code %d - %s" %\
                    (response.status, response.reason))
        rawmsg = response.read()
        self._build_answer(rawmsg)
        return expected_answer

    # No subscription for the http client
    def subscribe(self, signal_name, callback): raise NotImplementedError
    def unsubscribe(self, sub_id): raise NotImplementedError

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import os
from os import path
import kaa.metadata
from twisted.internet import reactor
from pytyxi import xine

from deejayd.player import PlayerError
from deejayd.player._base import *
from deejayd.ui import log


class XinePlayer(UnknownPlayer):
    name = "xine"
    supported_extensions = None
    xine_plugins = None

    def __init__(self, db, plugin_manager, config):
        UnknownPlayer.__init__(self, db, plugin_manager, config)
        self.__xine_options = {
            "video": self.config.get("xine", "video_output"),
            "display" : self.config.get("xine", "video_display"),
            "osd_support" : self.config.getboolean("xine", "osd_support"),
            "osd_font_size" : self.config.getint("xine", "osd_font_size"),
            "software_mixer": self.config.getboolean("xine", "software_mixer"),
            }
        self.__video_aspects = {
                "auto": xine.Stream.XINE_VO_ASPECT_AUTO,
                "1:1": xine.Stream.XINE_VO_ASPECT_SQUARE,
                "4:3": xine.Stream.XINE_VO_ASPECT_4_3,
                "16:9": xine.Stream.XINE_VO_ASPECT_ANAMORPHIC,
                "2.11:1": xine.Stream.XINE_VO_ASPECT_DVB,
            }
        self.__default_aspect_ratio = "auto"

        # init main instance
        try:
            self.__xine = xine.XinePlayer()
        except xine.XineError:
            raise PlayerError(_("Unable to init a xine instance"))

        # init vars
        self.__supports_gapless = self.__xine.has_gapless()

        self.__audio_volume = 100
        self.__video_volume = 100

        self.__window = None
        self.__stream = None
        self.__osd = None

    def start_play(self):
        super(XinePlayer, self).start_play()
        if not self._media_file: return

        # format correctly the uri
        uri = self._media_file["uri"].encode("utf-8")
        # For dvd chapter
        if "chapter" in self._media_file.keys() and \
                    self._media_file["chapter"] != -1:
            uri += ".%d" % self._media_file["chapter"]
        # load subtitle
        if "external_subtitle" in self._media_file and \
                self._media_file["external_subtitle"].startswith("file://"):
            # external subtitle
            uri += "#subtitle:%s" \
                    % self._media_file["external_subtitle"].encode("utf-8")
            self._media_file["subtitle"] = [{"lang": "none", "ix": -2},\
                                            {"lang": "auto", "ix": -1},\
                                            {"lang": "external", "ix":0}]
        elif "subtitle_channels" in self._media_file.keys() and\
                int(self._media_file["subtitle_channels"]) > 0:
            self._media_file["subtitle"] = [{"lang": "none", "ix": -2},\
                    {"lang": "auto", "ix": -1}]
            for i in range(int(self._media_file["subtitle_channels"])):
                self._media_file["subtitle"].append(\
                    {"lang": _("Sub channel %d") % (i+1,), "ix": i})
        # audio channels
        if "audio_channels" in self._media_file.keys() and \
                int(self._media_file["audio_channels"]) > 1:
            audio_channels = [{"lang":"none","ix":-2},{"lang":"auto","ix":-1}]
            for i in range(int(self._media_file["audio_channels"])):
                audio_channels.append(\
                        {"lang": _("Audio channel %d") % (i+1,), "ix": i})
            self._media_file["audio"] = audio_channels

        needs_video = self.current_is_video()
        if self.__stream:
            stream_should_change = (needs_video and\
                                    not self.__stream.has_video())\
                                or (not needs_video and
                                    self.__stream.has_video())
        else:
            stream_should_change = True
        if stream_should_change:
            self._create_stream(needs_video)

        def open_uri(uri):
            try:
                self.__stream.open(uri)
                self.__stream.play(0, 0)
            except xine.XineError:
                msg = _("Unable to play file %s") % uri
                log.err(msg)
                raise PlayerError(msg)

        if self._media_file["type"] == "webradio":
            while True:
                try: open_uri(self._media_file["uri"])
                except PlayerError, ex:
                    if self._media_file["url-index"] < \
                                            len(self._media_file["urls"])-1:
                        self._media_file["url-index"] += 1
                        self._media_file["uri"] = \
                                self._media_file["urls"]\
                                [self._media_file["url-index"]].encode("utf-8")
                    else:
                        raise ex
                else:
                    break
        else:
            try: open_uri(uri)
            except PlayerError, ex:
                self._destroy_stream()
                raise ex

        if self.__window:
            self.__window.show(self.current_is_video())

        # init video information
        if needs_video:
            self._media_file["av_offset"] = 0
            self._media_file["zoom"] = 100
            if "audio" in self._media_file:
                self._media_file["audio_idx"] = self.__stream.get_param(\
                        xine.Stream.XINE_PARAM_AUDIO_CHANNEL_LOGICAL)
            else:
                self._player_set_alang(-1) # auto audio channel
            if "subtitle" in self._media_file:
                self._media_file["sub_offset"] = 0
                self._media_file["subtitle_idx"] = self.__stream.get_param(\
                        xine.Stream.XINE_PARAM_SPU_CHANNEL)
            else:
                self._player_set_slang(-1) # auto subtitle channel
                try: del self._media_file["sub_offset"]
                except KeyError: pass
            # set video aspect ration to default value
            self.set_aspectratio(self.__default_aspect_ratio)

    def _change_file(self, new_file, gapless = False):
        sig = self.get_state() == PLAYER_STOP and True or False
        if self._media_file == None\
                or new_file == None\
                or self._media_file["type"] != new_file["type"]:
            self._destroy_stream()
            gapless = False

        self._media_file = new_file
        if gapless and self.__supports_gapless:
            self.__stream.set_param(xine.Stream.XINE_PARAM_GAPLESS_SWITCH, 1)
        self.start_play()
        if gapless and self.__supports_gapless:
            self.__stream.set_param(xine.Stream.XINE_PARAM_GAPLESS_SWITCH, 0)

        # replaygain reset
        self.set_volume(self.get_volume(), sig=False)

        if sig: self.dispatch_signame('player.status')
        self.dispatch_signame('player.current')

    def pause(self):
        if self.get_state() == PLAYER_PAUSE:
            self.__stream.set_param(xine.Stream.XINE_PARAM_SPEED,
                                    xine.Stream.XINE_SPEED_NORMAL)
        elif self.get_state() == PLAYER_PLAY:
            self.__stream.set_param(xine.Stream.XINE_PARAM_SPEED,
                                    xine.Stream.XINE_SPEED_PAUSE)
        else: return
        self.dispatch_signame('player.status')

    def stop(self):
        if self.get_state() != PLAYER_STOP:
            self._source.queue_reset()
            self._change_file(None)
            self.dispatch_signame('player.status')

    def set_zoom(self, zoom):
        if zoom > xine.Stream.XINE_VO_ZOOM_MAX\
        or zoom < xine.Stream.XINE_VO_ZOOM_MIN:
            raise PlayerError(_("Zoom value not accepted"))
        self.__stream.set_param(xine.Stream.XINE_PARAM_VO_ZOOM_X, zoom)
        self.__stream.set_param(xine.Stream.XINE_PARAM_VO_ZOOM_Y, zoom)
        self._media_file["zoom"] = zoom
        self._osd_set(_("Zoom: %d percent") % zoom)

    def set_aspectratio(self, aspect_ratio):
        try: asp = self.__video_aspects[aspect_ratio]
        except KeyError:
            raise PlayerError(_("Video aspect ration %s is not known.")\
                    % aspect_ratio)
        self.__default_aspect_ratio = aspect_ratio
        self._media_file["aspect_ratio"] = self.__default_aspect_ratio
        if self.__stream.has_video():
            self.__stream.set_param(xine.Stream.XINE_PARAM_VO_ASPECT_RATIO, asp)

    def set_avoffset(self, offset):
        self.__stream.set_param(xine.Stream.XINE_PARAM_AV_OFFSET, offset * 90)
        self._media_file["av_offset"] = offset
        self._osd_set(_("Audio/Video offset: %d ms") % offset)

    def set_suboffset(self, offset):
        if "subtitle" in self._media_file.keys():
            self.__stream.set_param(xine.Stream.XINE_PARAM_SPU_OFFSET,
                                    offset * 90)
            self._media_file["sub_offset"] = offset
            self._osd_set(_("Subtitle offset: %d ms") % offset)

    def _player_set_alang(self,lang_idx):
        self.__stream.set_param(xine.Stream.XINE_PARAM_AUDIO_CHANNEL_LOGICAL,
                                lang_idx)

    def _player_set_slang(self,lang_idx):
        self.__stream.set_param(xine.Stream.XINE_PARAM_SPU_CHANNEL, lang_idx)

    def _player_get_alang(self):
        return self.__stream.get_param(xine.Stream.\
                XINE_PARAM_AUDIO_CHANNEL_LOGICAL)

    def _player_get_slang(self):
        return self.__stream.get_param(xine.Stream.XINE_PARAM_SPU_CHANNEL)

    def get_volume(self):
        if self.current_is_video():
            return self.__video_volume
        else:
            return self.__audio_volume

    def set_volume(self, vol, sig = True):
        new_volume = min(100, int(vol))
        if self.current_is_video():
            self.__video_volume = new_volume
        else:
            self.__audio_volume = new_volume

        # replaygain support
        vol = self.get_volume()
        if self._replaygain and self._media_file is not None:
            try: scale = self._media_file.replay_gain()
            except AttributeError: pass # replaygain not supported
            else:
                vol = max(0.0, min(4.0, float(vol)/100.0 * scale))
                vol = min(100, int(vol * 100))
        if self.__stream:
            self.__stream.set_volume(vol)
        if sig:
            self._osd_set("Volume: %d" % self.get_volume())
            self.dispatch_signame('player.status')

    def get_position(self):
        if not self.__stream: return 0
        return self.__stream.get_pos()

    def _set_position(self,pos):
        pos = int(pos * 1000)
        state = self.get_state()
        if state == PLAYER_PAUSE:
            self.__stream.play(0, pos)
            self.__stream.set_param(xine.Stream.XINE_PARAM_SPEED,
                                    xine.Stream.XINE_SPEED_PAUSE)
        elif state == PLAYER_PLAY:
            self.__stream.play(0, pos)
        self.dispatch_signame('player.status')

    def get_state(self):
        if not self.__stream: return PLAYER_STOP

        status = self.__stream.get_status()
        if status == xine.Stream.XINE_STATUS_PLAY:
            if self.__stream.get_param(xine.Stream.XINE_PARAM_SPEED)\
               == xine.Stream.XINE_SPEED_NORMAL:
                return PLAYER_PLAY
            return PLAYER_PAUSE
        return PLAYER_STOP

    def is_supported_uri(self,uri_type):
        if self.xine_plugins == None:
            self.xine_plugins = self.__xine.list_input_plugins()

        return uri_type in self.xine_plugins

    def is_supported_format(self,format):
        if self.supported_extensions == None:
            self.supported_extensions = self.__xine.get_supported_extensions()
        return format.strip(".") in self.supported_extensions

    def current_is_video(self):
        return self._media_file is not None\
               and self._media_file['type'] == 'video'

    def close(self):
        UnknownPlayer.close(self)
        self.__xine.destroy()

    def _create_stream(self, has_video = True):
        if self.__stream != None:
            self._destroy_stream()

        # open audio driver
        driver_name = self.config.get("xine", "audio_output")
        try:
            audio_port = xine.AudioDriver(self.__xine, driver_name)
        except xine.xineError:
            raise PlayerError(_("Unable to open audio driver"))

        # open video driver
        if has_video and self._video_support\
                 and self.__xine_options["video"] != "none":
            try:
                video_port = xine.VideoDriver(self.__xine,
                                              self.__xine_options["video"],
                                              self.__xine_options["display"],
                                              self._fullscreen)
            except xine.XineError:
                msg = _("Unable to open video driver")
                log.err(msg)
                raise PlayerError(msg)
            else:
                self.__window = video_port.window
        else:
            video_port = None

        # create stream
        self.__stream = self.__xine.stream_new(audio_port, video_port)
        self.__stream.set_software_mixer(self.__xine_options["software_mixer"])

        if video_port and self.__xine_options["osd_support"]:
            self.__osd = self.__stream.osd_new(\
                                       self.__xine_options["osd_font_size"])

        # add event listener
        self.__stream.add_event_callback(self._event_callback)

        # restore volume
        self.__stream.set_volume(self.get_volume())

    def _destroy_stream(self):
        if self.__stream:
            self.__stream.destroy()
            self.__stream = None
            self.__window = None
            self.__osd = None

    def _osd_set(self, text):
        if not self.__osd: return
        self.__osd.clear()
        self.__osd.draw_text(60, 20, text.encode("utf-8"))
        self.__osd.show()

        reactor.callLater(2, self._osd_hide, text)

    def _osd_hide(self, text):
        if self.__osd:
            self.__osd.hide(text)

    #
    # callbacks
    #
    def _eof(self):
        if self._media_file:
            if self._media_file["type"] == "webradio":
                # an error happened, try the next url
                if self._media_file["url-index"] \
                        < len(self._media_file["urls"])-1:
                    self._media_file["url-index"] += 1
                    self._media_file["uri"] = \
                            self._media_file["urls"]\
                                [self._media_file["url-index"]].encode("utf-8")
                    self.start_play()
                return False
            else:
                try: self._media_file.played()
                except AttributeError: pass
                for plugin in self.plugins:
                    plugin.on_media_played(self._media_file)
        new_file = self._source.next(explicit = False)
        try: self._change_file(new_file, gapless = True)
        except PlayerError:
            pass
        return False

    def _update_metadata(self):
        if not self._media_file or self._media_file["type"] != "webradio":
            return False

        # update webradio song info
        meta = [
            (xine.Stream.XINE_META_INFO_TITLE, 'song-title'),
            (xine.Stream.XINE_META_INFO_ARTIST, 'song-artist'),
            (xine.Stream.XINE_META_INFO_ALBUM, 'song-album'),
        ]
        for info, name in meta:
            text = self.__stream.get_meta_info(info)
            if not text:
                continue
            text = text.decode('UTF-8', 'replace')
            if name not in self._media_file.keys() or\
                           self._media_file[name] != text:
                self._media_file[name] = text
        self.dispatch_signame('player.current')
        return False

    # this callback is not called in the main reactor thread
    # so we have to use callFromThread function instead of callLater
    # see |http://twistedmatrix.com/documents/current/api/
    #     |twisted.internet.interfaces.IReactorThreads.callFromThread.html
    def _event_callback(self, user_data, event):
        if event.type == xine.Event.XINE_EVENT_UI_PLAYBACK_FINISHED:
            log.info("Xine event : playback finished")
            reactor.callFromThread(self._eof)
        elif event.type == xine.Event.XINE_EVENT_UI_SET_TITLE:
            log.info("Xine event : set title")
            reactor.callFromThread(self._update_metadata)
        elif event.type == xine.Event.XINE_EVENT_UI_MESSAGE:
            log.info("Xine event : message")
            try:
                message = event.message()
            except xine.XineError, errornum:
                message = _("Xine error %s") % errornum
            if message is not None:
                reactor.callFromThread(log.err, message)
        return True


class DvdParser:
    DEVICE = "/dev/dvd"

    def __init__(self):
        try:
            self.__xine = xine.XinePlayer()
        except xine.XineError:
            raise PlayerError(_("Unable to init a xine instance"))
        self.__mine_stream = self.__xine.stream_new(None, None)

    def get_dvd_info(self):
        kaa_infos = kaa.metadata.parse(self.DEVICE)
        if kaa_infos is None:
            raise PlayerError(_("Unable to identify dvd device"))

        dvd_info = {"title": kaa_infos["label"], 'track': []}
        longest_track = {"ix": 0, "length": 0}
        for idx, t in enumerate(kaa_infos['tracks']):
            try: self.__mine_stream.open("dvd://%d" % (idx+1,))
            except xine.XineError, ex:
                raise PlayerError, ex
            track = {"ix": idx+1, "length": int(t['length']), "chapter": []}
            if track['length'] > longest_track["length"]:
                longest_track = track

            # get audio channels info
            channels_number = len(t['audio'])
            audio_channels = [{"lang":"none","ix":-2},{"lang":"auto","ix":-1}]
            for ch in range(0,channels_number):
                lang = self.__mine_stream.get_audio_lang(ch)
                audio_channels.append({'ix':ch, "lang":lang.encode("utf-8")})
            track["audio"] = audio_channels

            # get subtitles channels info
            channels_number = len(t['subtitles'])
            sub_channels = [{"lang":"none","ix":-2},{"lang":"auto","ix":-1}]
            for ch in range(0,channels_number):
                lang = self.__mine_stream.get_spu_lang(ch)
                sub_channels.append({'ix':ch, "lang":lang.encode("utf-8")})
            track["subp"] = sub_channels

            # chapters
            for c_i,chapter in enumerate(kaa_infos['tracks'][idx]['chapters']):
                track["chapter"].append({ "ix": c_i+1,\
                        'length': int(chapter["pos"]) })
                if c_i > 0:
                    track["chapter"][c_i-1]['length'] = int(chapter["pos"]) - \
                            track["chapter"][c_i-1]['length']
                if c_i == len(kaa_infos['tracks'][idx]['chapters']) - 1:
                    track["chapter"][c_i]['length'] = track["length"] - \
                            int(track["chapter"][c_i]['length'])

            dvd_info['track'].append(track)

        dvd_info['longest_track'] = longest_track["ix"]
        return dvd_info

    def close(self):
        self.__mine_stream.destroy()
        self.__xine.destroy()

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from deejayd.mediadb.formats._base import _AudioFile

extensions = [".ogg"]
try: from mutagen.oggvorbis import OggVorbis
except ImportError:
    extensions = []

class OggFile(_AudioFile):
    _tagclass_ = OggVorbis

    def get_cover(self, tag_info):
        return None
        # disable for now
        # not work correctly
        #if 'coverarttype' in tag_info.keys() and\
        #        int(tag_info['coverarttype'][0])==3:
        #    try:
        #        return {"data": tag_info['coverart'][0],\
        #                "mime": tag_info['coverartmime'][0]}
        #    except KeyError:
        #        return None
        #return None

object = OggFile

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import random

__all__ = ["SimpleMediaList", "SortedMediaList", "UnsortedMediaList"]

class SimpleMediaList(object):

    def __init__(self, list_id = 0):
        self._media_id = 0
        self._order, self._content = [], {}
        self.repeat = False

        self.list_id = list_id
        self.time_length = 0

    def __len__(self):
        return len(self._order)

    def _set_media_id(self):
        self._media_id += 1
        return self._media_id

    def _set_media_ans(self, id, pos = -1):
        try: ans = self._content[id]
        except KeyError:
            return None
        ans.update({"id": id, "pos": pos})
        return ans

    def get(self, start = 0, stop = None):
        stop = stop or len(self._order)
        return map(self._set_media_ans, self._order[start:stop],\
                    range(start, stop))

    def get_item(self, id):
        try: return self._set_media_ans(id, self._order.index(id))
        except ValueError:
            return None

    def get_item_first(self):
        if self._order:
            return self._set_media_ans(self._order[0], 0)
        return None

    def get_item_last(self):
        if self._order:
            return self._set_media_ans(self._order[-1], len(self._order)-1)
        return None

    def get_ids(self):
        return self._order

    def reload_item_pos(self, media):
        try: pos = self._order.index(media["id"])
        except TypeError: return
        except ValueError:
            pos = -1
        media["pos"] = pos

    def set(self, medias):
        self._order, self._content = [], {}
        self.time_length = 0
        self.add_media(medias)

    def add_media(self, medias, first_pos = None):
        first_pos = first_pos or len(self._order)

        for index, m in enumerate(medias):
            id = self._set_media_id()
            self._order.insert(first_pos + index, id)
            self._content[id] = m
            # update medialist time length
            try: length = int(m["length"])
            except (ValueError, KeyError, TypeError):
                continue
            self.time_length += length
        self.list_id += 1

    def clear(self):
        self._order, self._content = [], {}
        self.time_length = 0
        self.list_id += 1

    def delete(self, ids, type = "id"):
        if type == "pos": ids = [self._order[p] for p in ids]
        missing_ids = [id for id in ids if id not in self._order]
        if missing_ids: return False

        for id in ids:
            self._order.remove(id)
            # update time length
            try: length = int(self._content[id]["length"])
            except (ValueError, KeyError, TypeError): continue
            else: self.time_length -= length
            del self._content[id]
        self.list_id += 1
        return True

    def next(self, media):
        try:
            idx = self._order.index(media["id"])
            id = self._order[idx+1]
        except ValueError: # id not found, try to find by media_id
            id = None
            if media:
                for list_id, m in self._content.items():
                    if m["media_id"] == media["media_id"]:
                        idx = self._order.index(m["id"])
                        try: id = self._order[idx+1]
                        except IndexError: # end of medialist
                            return None
                        break
            if id is None: return None
        except (IndexError, KeyError, TypeError):
            return None
        return self._set_media_ans(id, idx+1)

    def previous(self, media):
        try:
            idx = self._order.index(media["id"])
            id = idx and self._order[idx-1] or None
        except (ValueError, KeyError):
            return None
        return self._set_media_ans(id, idx-1)

    def find_id(self, media_id):
        for id, m in self._content.items():
            if m["media_id"] == media_id:
                return id
        raise ValueError


class _MediaList(SimpleMediaList):

    #
    # library changes action
    #
    def remove_media(self, media_id):
        ans = False
        for key, media in self._content.items():
            if media["media_id"] == media_id:
                self._order.remove(key)
                del self._content[key]
                ans = True
                try: length = int(media["length"])
                except (ValueError, KeyError, TypeError):
                    continue
                self.time_length -= length
        if ans: self.list_id += 1
        return ans

    def update_media(self, new_media):
        ans = False
        for key, m in self._content.items():
            if m["media_id"] == new_media["media_id"]:
                self._content[key] = new_media
                ans = True
        if ans: self.list_id += 1
        return ans


class UnsortedMediaList(_MediaList):

    def move(self, ids, new_pos, type = "id"):
        if type == "pos": ids = [self._order[p] for p in ids]
        missing_ids = [id for id in ids if id not in self._order]
        if missing_ids: return False

        s_list = [id for id in self._order[:new_pos] if id not in ids]
        e_list = [id for id in self._order[new_pos:] if id not in ids]
        self._order = s_list + ids + e_list
        self.list_id += 1
        return True

    def shuffle(self, current = None):
        if not self._order: return
        random.shuffle(self._order)
        if current and current["id"]:
            try: self._order.remove(current["id"])
            except ValueError: pass
            else:
                self._order = [current["id"]] + self._order
        self.list_id += 1


class SortedMediaList(_MediaList):

    def __init__(self, list_id = 0):
        super(SortedMediaList, self).__init__(list_id)

    #
    # sort actions
    #
    def __compare_tag(self, id1, id2, tag, direction):
        m1 = self._content[id1]
        m2 = self._content[id2]
        if m1[tag] < m2[tag]:
            return direction == "ascending" and -1 or 1
        elif m1[tag] == m2[tag]:
            return 0
        else:
            return direction == "ascending" and 1 or -1

    def sort(self, sorts):

        def compare(id1, id2):
            for (tag, direction) in sorts:
                result = self.__compare_tag(id1, id2, tag, direction)
                if result != 0: return result
            return 0

        self._order.sort(cmp=compare)
        self.list_id += 1

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from deejayd.mediadb.formats._base import _AudioFile

extensions = ['.mp4', '.m4a']
try: from mutagen.mp4 import MP4
except ImportError:
    extensions = []

class Mp4File(_AudioFile):
    __translate = {
        "\xa9nam": "title",
        "\xa9alb": "album",
        "\xa9ART": "artist",
        "\xa9day": "date",
        "\xa9gen": "genre",
        "----:com.apple.iTunes:replaygain_track_gain": "replaygain_track_gain",
        "----:com.apple.iTunes:replaygain_track_peak": "replaygain_track_peak",
        }
    __tupletranslate = {
        "trkn": "tracknumber",
        "disk": "discnumber",
        }

    def parse(self, file):
        infos = _AudioFile.parse(self, file)
        mp4_info = MP4(file)
        infos["bitrate"] = int(mp4_info.info.bitrate)
        infos["length"] = int(mp4_info.info.length)

        for tag, name in self.__translate.iteritems():
            try: infos[name] = mp4_info[tag][0]
            except: infos[name] = '';

        for tag, name in self.__tupletranslate.iteritems():
            try:
                cur, total = mp4_info[tag][0]
                if total: self[name] = "%02d/%02d" % (cur, total)
                else: infos[name] = "%02d" % cur
            except: infos[name] = '';

        # extract cover
        try: cover = mp4_info["covr"][0]
        except (KeyError, ValueError):
            pass
        else:
            mime = "image/jpeg"
            if cover.format == cover.FORMAT_PNG:
                mime = "image/png"
            infos["cover"] = {"mime": mime, "data": cover}

        infos["various_artist"] = infos["artist"]
        return infos

object = Mp4File

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import ConfigParser, os, sys, string

class DeejaydConfig:

    custom_conf = None
    __global_conf = '/etc/deejayd.conf'
    __user_conf = '~/.deejayd.conf'
    __config = None

    def __init__(self):

        if DeejaydConfig.__config == None:
            DeejaydConfig.__config = ConfigParser.SafeConfigParser()

            default_config_path = os.path.abspath(os.path.dirname(__file__))
            DeejaydConfig.__config.readfp(open(default_config_path\
                                        + '/defaults.conf'))

            conf_files = [DeejaydConfig.__global_conf,\
                         os.path.expanduser(DeejaydConfig.__user_conf)]
            if DeejaydConfig.custom_conf:
                conf_files.append(DeejaydConfig.custom_conf)
            DeejaydConfig.__config.read(conf_files)

    def __getattr__(self, name):
        return getattr(DeejaydConfig.__config, name)

    def set(self, section, variable, value):
        self.__config.set(section, variable, value)

    def getlist(self, section, variable):
        list_items = self.__config.get(section, variable).split(',')
        return map(string.strip, list_items)

    def get_bind_addresses(self, service = 'net'):
        bind_addresses = self.getlist(service, 'bind_addresses')
        if 'all' in bind_addresses:
            return ['']
        else:
            return bind_addresses

    def write(self, fp):
        self.__config.write(fp)

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# need python-lxml
try: from lxml import etree as ET
except ImportError:
    import sys
    msg = "Unable to import module to build/parse xml, please install lxml"
    sys.exit(_(msg))

class DeejaydXMLObject(object):

    def __is_xml_char(self, char):
        # FIXME: This function should probably be more complicated than this.
        return char not in '\x19'

    def _to_xml_string(self, s):
        xml_string = ''
        if isinstance(s, int) or isinstance(s, float) or isinstance(s, long):
            xml_string = "%d" % (s,)
        elif isinstance(s, str):
            xml_string = "%s" % (s.decode('utf-8'))
        elif isinstance(s, unicode):
            rs = s.encode("utf-8")
            xml_string = "%s" % (rs.decode('utf-8'))
        else:
            raise TypeError

        filtered_xml_string = []
        for string_char in xml_string:
            if self.__is_xml_char(string_char):
                filtered_xml_string.append(string_char)
        return ''.join(filtered_xml_string)

    def _indent(self,elem, level=0):
        indent_char = "    "
        i = "\n" + level*indent_char
        if len(elem):
            if not elem.text or not elem.text.strip():
                elem.text = i + indent_char
            for e in elem:
                self._indent(e, level+1)
                if not e.tail or not e.tail.strip():
                    e.tail = i + indent_char
            if not e.tail or not e.tail.strip():
                e.tail = i
        else:
            if level and (not elem.tail or not elem.tail.strip()):
                elem.tail = i

    def to_string(self):
        return ET.tostring(self.xmlroot)

    def to_xml(self):
        return '<?xml version="1.0" encoding="utf-8"?>' + self.to_string()

    def to_pretty_xml(self):
        self._indent(self.xmlroot)
        return '<?xml version="1.0" encoding="utf-8"?>' + "\n" +\
            self.to_string() + "\n"

# vim: ts=4 sw=4 expandtab
